#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Dec  3 18:13:45 2024

@author: ncro8394
"""


from os import listdir, mkdir, path
import shutil
from wonambi import Dataset, graphoelement
from wonambi.attr import Annotations, create_empty_annotations
from wonambi.detect import DetectSlowWave
from wonambi.trans import fetch
import mne
from scipy.signal import find_peaks, peak_widths
import yasa
from xml.etree.ElementTree import Element, SubElement, tostring, parse
from numpy import array, multiply, nan
from copy import deepcopy
from datetime import datetime, date
from pandas import DataFrame
from ..utils.logs import create_logger, create_logger_outfile
from ..utils.load import (load_channels, load_sessions, load_stagechan, load_emg, 
                          load_eog, read_inversion, rename_channels)
from ..utils.misc import remove_duplicate_evts
 

class SAND:
    
    """ Seapipe Artefact and Noise Detection (S.A.N.D)

        This module runs automated artefact detection with the option of using
        previously published staging algorithms:
            1. YASA (standard deviation)
            2. YASA (covariance)
            3. (More to come..)
        
    """   
    
    def __init__(self, rec_dir, xml_dir, out_dir, log_dir, eeg_chan, ref_chan,
                 eog_chan, emg_chan, rater = None, subs='all', sessions='all', 
                 tracking = None):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.log_dir = log_dir
        
        self.eeg_chan = eeg_chan
        self.ref_chan = ref_chan
        self.eog_chan = eog_chan
        self.emg_chan = emg_chan
        self.rater = rater
        
        self.subs = subs
        self.sessions = sessions
        
        if tracking == None:
            tracking = {}
        self.tracking = tracking


    def detect_artefacts(self, method, invert = False, filetype = '.edf', 
                               win_size = 5,
                               logger = create_logger('Detect artefacts')):
        
        ''' Automatically detects sleep stages by applying a published 
            prediction algorithm.
        
            Creates a new annotations file if one doesn't already exist.
        
        INPUTS:
            
            method      ->   str of name of automated detection algorithm to 
                             detect staging with. Currently only 'Vallat2021' 
                             is supported. 
                             (https://doi.org/10.7554/eLife.70092)
                             
            qual_thresh ->   Quality threshold. Any stages with a confidence of 
                             prediction lower than this threshold will be set 
                             to 'Undefined' for futher manual review.
   
        
        '''
        
        ### 0.a Set up logging
        flag = 0
        tracking = self.tracking
        
        logger.info('')
        logger.debug(rf"""Commencing artefact detection... 
                     
                                             ____
                                      /^\   / -- )
                                     / | \ (____/
                                    / | | \ / /
                                   /_|_|_|_/ /
                                    |     / /
                     __    __    __ |    / /__    __    __
                    [  ]__[  ]__[  ].   / /[  ]__[  ]__[  ]     ......
                    |__            ____/ /___           __|    .......
                       |          / .------  )         |     ..........
                       |         / /        /          |    ............
                       |        / /        / _         |  ...............
                   ~._..-~._,….-ˆ‘ˆ˝\_,~._;––' \_.~.~._.~'\................  
                       
            
                    Seapipe Artefact and Noise Detection
                    (S.A.N.D)

                    
                                                    """,)
        ### 1. First we check the directories
        # a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug("Output directory: " + self.out_dir + " exists")
        else:
            mkdir(self.out_dir)
        
        # b. Check input list
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [p for p in subs if not '.' in p]
        else:
            logger.error("'subs' must either be an array of subject ids or = 'all' ")  
            return
        
        ### 2. Begin loop through dataset
       
        # a. Begin loop through participants
        subs.sort()
        for i, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            # b. Begin loop through sessions
            flag, sessions = load_sessions(sub, self.sessions, self.rec_dir, flag, 
                                     logger, verbose=2)
            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'slowosc':{}} 
                
                # Load chans
                pflag = deepcopy(flag)
                flag, chanset = load_channels(sub, ses, self.eeg_chan, self.ref_chan,
                                              flag, logger)
                if flag - pflag > 0:
                    logger.warning(f'Skipping {sub}, {ses}...')
                    break
                eeg_chan = [x for x in chanset]
                ref_chan = chanset[eeg_chan[0]]
                flag, emg_chan = load_emg(sub, ses, self.eeg_chan, 
                                    flag, logger)
                flag, eog_chan = load_eog(sub, ses, self.eeg_chan,
                                    flag, logger)
                if not isinstance(eeg_chan, list):
                    eeg_chan = [eeg_chan]
                if not isinstance(ref_chan, list):
                    ref_chan = [ref_chan]
                if not isinstance(eog_chan, list):
                    eog_chan = [eog_chan]
                if not isinstance(emg_chan, list):
                    emg_chan = [emg_chan]
                
                ## c. Load recording
                rdir = self.rec_dir + '/' + sub + '/' + ses + '/eeg/'
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    chans = eeg_chan + ref_chan + eog_chan + emg_chan
                    chans = [x for x in chans if x]
                    raw = mne.io.read_raw_edf(rdir + edf_file[0], 
                                              include = chans,
                                              preload=True, verbose = False)
                except:
                    logger.warning(f' No input {filetype} file in {rdir}')
                    flag += 1
                    break
                    
                # d. Load/create for annotations file
                if not path.exists(self.xml_dir + '/' + sub):
                    mkdir(self.xml_dir + '/' + sub)
                if not path.exists(self.xml_dir + '/' + sub + '/' + ses):
                     mkdir(self.xml_dir + '/' + sub + '/' + ses)
                xdir = self.xml_dir + '/' + sub + '/' + ses
                xml_file = f'{xdir}/{sub}_{ses}_eeg.xml'
                if not path.exists(xml_file):
                    dset = Dataset(rdir + edf_file[0])
                    create_empty_annotations(xml_file, dset)
                    logger.warning(f'No annotations file exists. Creating annotations file for {sub}, {ses} and detecting Artefacts WITHOUT hypnogram.')
                    annot = Annotations(xml_file)
                    hypno = None
                else:
                    logger.debug(f'Annotations file exists for {sub}, {ses}, staging will be used for Artefact detection.')

                    # Extract hypnogram
                    annot = Annotations(xml_file)
                    hypno = [x['stage'] for x in annot.get_epochs()]
                    stage_key = {'Wake':0,
                                 'NREM1':1,
                                 'NREM2':2,
                                 'NREM3':3,
                                 'REM':4,
                                 'Undefined':0,
                                 'Unknown':0}
                    
                    hypno = array([int(stage_key[x]) for x in hypno])
                    sf = raw.info["sfreq"]
                
                if 'yasa' in method:
                    if 'covar' in method:
                        yasa_meth = 'covar'
                    else:
                        yasa_meth = 'std'
                        
                    # Convert raw data to array    
                    data = raw.to_data_frame()
                    inds = [x for x in data if x in eeg_chan]
                    data = data[inds].T
                    data = data.to_numpy()
                    
                    # Upsample hypnogram to match raw data
                    hypno_up = yasa.hypno_upsample_to_data(hypno, 1/30, data, 
                                                           sf_data=sf)
                    
                    # Detect artefacts
                    art, zscores = yasa.art_detect(data, sf, window = win_size, 
                                                           hypno = hypno_up, 
                                                           include = (1, 2, 3, 4), 
                                                           method = yasa_meth, 
                                                           threshold = 3, 
                                                           n_chan_reject = 2, 
                                                           verbose = False)
                    
                    # Upsample artefacts to match raw data
                    art = multiply(art, 1)
                    sf_art = 1/win_size
                    art_up = yasa.hypno_upsample_to_data(art, sf_art, data, sf)
                    
                    # Find start/end times of artefacts
                    peaks = find_peaks(art_up)
                    properties = peak_widths(art_up, peaks[0])
                    times = [x for x in zip(properties[2],properties[3])]
    
                    # Convert to wonambi annotations format
                    evts = []
                    for x in times:
                        evts.append({'name':'Artefact',
                              'start':x[0]/sf,
                              'end':x[1]/sf,
                              'chan':[''],
                              'stage':'',
                              'quality':'Good',
                              'cycle':''})
                        
                    # Add to annotations file
                    grapho = graphoelement.Graphoelement()
                    grapho.events = evts          
                    grapho.to_annot(annot)

                else:
                    logger.error("Currently the only method that is functioning is 'yasa_std' or 'yasa_covar.")
                
                # ### get cycles
                # if self.cycle_idx is not None:
                #     all_cycles = annot.get_cycles()
                #     cycle = [all_cycles[i - 1] for i in self.cycle_idx if i <= len(all_cycles)]
                # else:
                #     cycle = None
                
                # ### if event channel only, specify event channels
                # # 4.d. Channel setup
                # flag, chanset = load_channels(sub, ses, self.chan, 
                #                               self.ref_chan, flag, logger)
                # if not chanset:
                #     flag+=1
                #     break
                # newchans = rename_channels(sub, ses, self.chan, logger)

                # # get segments
                # for c, ch in enumerate(chanset):
                #     logger.debug(f"Reading data for {ch}:{'/'.join(chanset[ch])}")
                #     segments = fetch(dset, annot, cat = cat,  
                #                      stage = self.stage, cycle=cycle,  
                #                      epoch = epoch_opts['epoch'], 
                #                      epoch_dur = epoch_opts['epoch_dur'], 
                #                      epoch_overlap = epoch_opts['epoch_overlap'], 
                #                      epoch_step = epoch_opts['epoch_step'], 
                #                      reject_epoch = epoch_opts['reject_epoch'], 
                #                      reject_artf = epoch_opts['reject_artf'],
                #                      min_dur = epoch_opts['min_dur'])
                    
        return


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 29 10:29:11 2021

@author: Nathan Cross
"""
from copy import deepcopy
from csv import reader
from fooof import FOOOF
from fooof.analysis import get_band_peak_fm
from glob import glob
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from numpy import (asarray, ceil, concatenate, empty, floor, mean, nan, 
                   ones, pi, sqrt, stack, sum, zeros)
from openpyxl import Workbook
from os import listdir, mkdir, path, walk
from pandas import DataFrame, ExcelWriter
from scipy.fftpack import next_fast_len
from wonambi import ChanFreq, Dataset
from wonambi.attr import Annotations
from wonambi.trans import (fetch, frequency, get_descriptives, export_freq, 
                           export_freq_band)
from ..utils.misc import bandpass_mne, laplacian_mne, notch_mne, notch_mne2
from ..utils.logs import create_logger
from ..utils.load import load_channels, rename_channels


def default_general_opts():
    general_opts = {'freq_full':True, 
                    'freq_band':True, 
                    'freq_plot':False,
                    'max_freq_plot':35,
                    'suffix':None,
                    'chan_grp_name':'eeg'}
    return general_opts

def default_frequency_opts():
    bands = [(0.25, 1.25),  #SO
             (0.5, 4),      #delta
             (4, 7.75),     #theta
             (8, 11),       #alpha
             (11.25, 16),   #sigma
             (16.25, 19),   #low beta
             (19.25, 35)]   #high beta
    frequency_opts = {'bands': bands,
                      'output':'spectraldensity', 
                      'scaling':'power', 
                      'sides':'one', 
                      'taper':'hann', 
                      'detrend':'linear', 
                      'n_fft': None, 
                      'fast_fft': False, 
                      'duration': 4,
                      'overlap': 0.5,
                      'step': None,
                      'centend':'mean',
                      'log_trans': True,
                      'halfbandwidth': 3,
                      'NW': None}
    return frequency_opts

def default_filter_opts():
    filter_opts = {'laplacian': False, 
                   'oREF': None, 
                   'lapchan': None, 
                   'laplacian_rename': False, 
                   'renames': None, 
                   'montage': 'standard_alphabetic',
                   'notch': True, 
                   'notch_freq': 50, 
                   'notch_harmonics': False,
                   'bandpass': False,
                   'highpass': None,
                   'lowpass': 50}
    return filter_opts

def default_epoch_opts():
    epoch_opts = {'epoch': None,
                  'reject_epoch': True,
                  'reject_artf': True,
                  'min_dur':0,
                  'epoch_dur': 30,
                  'epoch_overlap': 0.5,
                  'epoch_step': None}
    
    return epoch_opts

def default_event_opts():
    event_opts = {'evt_type': None,
                  'event_chan': None}
    return event_opts

def default_fooof_opts():
    fooof_opts = {'psd_dur': 2, 
                  'peak_width_limits': [1, 12], 
                  'max_n_peaks': 8,
                  'min_peak_amplitude': 0.0,
                  'peak_threshold': 2.0,
                  'freq_range': [0.5, 35],
                  'select_highest': True,
                  'thresh_param': 'PW',
                  'bands_fooof': None, 
                  'thresh_select': None}
    return fooof_opts

def default_norm_opts():
    norm_opts = {'norm_cat': (1, 1, 1, 0),
                 'norm_evt_type': ['norm'], 
                 'norm_stage': None, 
                 'norm_epoch': None}
    return norm_opts


class Spectrum:
    
    """Design power spectral analyses on segmented data.

    Parameters
    ----------
    method : str
        one of the predefined methods
    frequency : tuple of float
        low and high frequency of frequency band
    duration : tuple of float
        min and max duration of spindles
    merge : bool
        if True, then after events are detected on every channel, events on 
        different channels that are separated by less than min_interval will be 
        merged into a single event, with 'chan' = the chan of the earlier-onset 
        event.
        
    Functions
    ----------
    fooof_it : 
        calculation of spectral parameters.
    powerspec_it : 
        Call to calculate power spectral analysis for each of:
            <sub>, <ses>, <chan>, <segment>
    powerspec_summary_full : 
        for power spectrum statistics. Returns value for whole spectrum, in 
        N/fs bins.
    powerspec_summary_bands : 
        for power spectrum statistics. Returns value for whole spectrum, 
        for pre-defined bands of interest.
        
    Notes
    -----
    See individual functions for other attribute descriptions.
    """ 
    
    def __init__(self, rec_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                 grp_name, stage, frequency = (11,16), rater = None, 
                 subs = 'all', sessions = 'all'):
        
        self.rec_dir = rec_dir
        self.xml_dir = xml_dir
        self.out_dir = out_dir
        self.log_dir = log_dir
        
        self.chan = chan
        self.ref_chan = ref_chan
        self.grp_name = grp_name
        self.stage = stage
        self.frequency = frequency
        self.rater = rater
        
        self.subs = subs
        self.sessions = sessions
        
        self.tracking = {}

    
    def fooof_it(self, general_opts, frequency_opts, filter_opts, 
                 epoch_opts, event_opts, fooof_opts, chan = None, rater = None, 
                 grp_name = 'eeg', cat = (1,1,1,1), cycle_idx = None,
                 filetype = '.edf', suffix = 'fooof'):
        

        '''
        FOOOF is a fast, efficient, and physiologically-informed tool to 
        parameterize neural power spectra.
        
        Outputs
        -------
        ap_params: 
                  Parameters for the aperiodic component, reflecting 1/f like 
                  characteristics, including: 
                                    - offset
                                    - slope
        pk_params:
                  Parameters for periodic components (putative oscillations), 
                  as peaks rising above the aperiodic component, including:
                                    - peak frequency (in Hz)
                                    - peak bandwidth
                                    - peak amplitude (in µV^2)
        
        Notes
        -----
        See: https://fooof-tools.github.io/fooof/index.html for more info.
        
        '''
        
        ### 0.a. Set up logging
        logger = create_logger('Specparam')
        tracking = self.tracking
        flag = 0
        
        logger.info('')
        
        ### 0.b. Set up organisation of export
        if cat[0] + cat[1] == 2:
            model = 'whole_night'
            logger.debug(f'Parameterizing power spectrum in range {self.frequency[0]}-{self.frequency[1]} Hz for the whole night.')
        elif cat[0] + cat[1] == 0:
            model = 'stage*cycle'
            logger.debug(f'Parameterizing power spectrum in range {self.frequency[0]}-{self.frequency[1]} Hz per stage and cycle separately.')
        elif cat[0] == 0:
            model = 'per_cycle'
            logger.debug(f'Parameterizing power spectrum in range {self.frequency[0]}-{self.frequency[1]} Hz per cycle separately.')
        elif cat[1] == 0:
            model = 'per_stage'  
            logger.debug(f'Parameterizing power spectrum in range {self.frequency[0]}-{self.frequency[1]} Hz per stage separately.')
        if 'cycle' in model and cycle_idx == None:
            logger.info('')
            logger.critical("To run cycles separately (i.e. cat[0] = 0), cycle_idx cannot be 'None'")
            return
        cat = tuple((cat[0],cat[1],1,1)) # force concatenation of discontinuous & events
        logger.info('')
        logger.debug(r"""
                     
                                  |
                                  |
                                  |  .
                                  |  \
                                  |   `~.
                        (µV2)     |      `~.       FOOOF !
                                  |         `^.
                                  |            `~.
                                  |               `•._
                                  |                   `~¬.…_
                                  |_____________________________
                                              (Hz)
                                  
                                                    """,)
        
        # 1.a. Define model parameters
        if fooof_opts is None:
            logger.critical('Options not set for FOOOF')
            return
        else:
            fm = FOOOF(fooof_opts['peak_width_limits'], fooof_opts['max_n_peaks'], 
                       fooof_opts['min_peak_amplitude'], fooof_opts['peak_threshold'])
            def gaussian_integral(a, c):
                """ Returns definite integral of a gaussian function with height a and
                standard deviation c."""
                return sqrt(2) * a * abs(c) * sqrt(pi)
        
        # 1.b. Prepare Bands
        if fooof_opts['bands_fooof'] is not None:
            bands = fooof_opts['bands_fooof']
        else:
            stp = min(fooof_opts['peak_width_limits'])
            low = int(floor(fooof_opts['freq_range'][0]))
            hi = int(ceil(fooof_opts['freq_range'][1]))
            bands = [(x,x + stp) for x in range(low,hi)]
        
        # 2.a. Check for output folder, if doesn't exist, create
        if path.exists(self.out_dir):
                logger.debug(f"Output directory: {self.out_dir} exists")
        else:
            mkdir(self.out_dir)
        
        # 2.b. Get subjects
        subs = self.subs
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = next(walk(self.xml_dir))[1]
        else:
            logger.info('')
            logger.critical("'subs' must either be an array of Participant IDs or = 'all' ")
            return
        subs.sort()
        
        # 3.a. Begin loop through participants
        for p, sub in enumerate(subs):
            tracking[f'{sub}'] = {}
            
            # b. Begin loop through sessions
            sessions = self.sessions
            if sessions == 'all':
                sessions = listdir(f'{self.rec_dir}/{sub}')
                sessions = [x for x in sessions if not '.' in x] 

            for v, ses in enumerate(sessions):
                logger.info('')
                logger.debug(f'Commencing {sub}, {ses}')
                tracking[f'{sub}'][f'{ses}'] = {'Spec_peaks':{}} 
                
                ## Define files
                rdir = f'{self.rec_dir}/{sub}/{ses}/eeg/'
                xdir = f'{self.xml_dir}/{sub}/{ses}/'
                
                try:
                    edf_file = [x for x in listdir(rdir) if x.endswith(filetype)]
                    dset = Dataset(rdir + edf_file[0])
                except:
                    logger.warning(f' No input {filetype} file in {rdir}. Skipping...')
                    break
                
                xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]            
                
                ### Define output path
                if not path.exists(self.out_dir):
                    mkdir(self.out_dir)
                if not path.exists(f'{self.out_dir}/{sub}'):
                    mkdir(f'{self.out_dir}/{sub}')
                if not path.exists(f'{self.out_dir}/{sub}/{ses}'):
                    mkdir(f'{self.out_dir}/{sub}/{ses}')
                outpath = f'{self.out_dir}/{sub}/{ses}'
                
                ## Now import data
                dset = Dataset(rdir + edf_file[0])
                annot = Annotations(xdir + xml_file[0], rater_name=rater)
                 
                ### get cycles
                if cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[i - 1] for i in cycle_idx if i <= len(all_cycles)]
                else:
                    cycle = None
                    
                ### if event channel only, specify event channels
                # 4.d. Channel setup
                if not chan:
                    flag, chanset = load_channels(sub, ses, self.chan, 
                                                  self.ref_chan, flag, logger)
                    if not chanset:
                        break
                
                    newchans = rename_channels(sub, ses, self.chan, logger)

                
                for c, ch in enumerate(chanset):
                    logger.debug(f"Reading data for {ch}:{'/'.join(chanset[ch])}")
                    segments = fetch(dset, annot, cat = cat, evt_type = None, 
                                     stage = self.stage, cycle=cycle,  
                                     epoch = epoch_opts['epoch'], 
                                     epoch_dur = epoch_opts['epoch_dur'], 
                                     epoch_overlap = epoch_opts['epoch_overlap'], 
                                     epoch_step = epoch_opts['epoch_step'], 
                                     reject_epoch = epoch_opts['reject_epoch'], 
                                     reject_artf = epoch_opts['reject_artf'],
                                     min_dur = epoch_opts['min_dur'])
                    
                    if len(segments)==0:
                        logger.warning(f"No valid data found for {ch}:{'/'.join(chanset[ch])}.")
                        continue
                    
                    # 5.b Rename channel for output file (if required)
                    if newchans:
                        fnamechan = newchans[ch]
                    else:
                        fnamechan = ch
                    
                    segments.read_data(chan = ch, ref_chan = chanset[ch])
                    
                    for sg, seg in enumerate(segments):
                        logger.debug(f'Analysing segment {sg+1} of {len(segments)}')
                        out = dict(seg)
                        data = seg['data']
                        timeline = data.axis['time'][0]
                        out['start'] = timeline[0]
                        out['end'] = timeline[-1]
                        out['duration'] = len(timeline) / data.s_freq
                        
                        if frequency_opts['fast_fft']:
                            n_fft = next_fast_len(data.number_of('time')[0])
                        else:
                            n_fft = frequency_opts['n_fft']
                    
                        Fooofxx = frequency(data, output=frequency_opts['output'], 
                                        scaling=frequency_opts['scaling'], 
                                        sides=frequency_opts['sides'], 
                                        taper=frequency_opts['taper'],
                                        halfbandwidth=frequency_opts['halfbandwidth'], 
                                        NW=frequency_opts['NW'],
                                        duration=frequency_opts['duration'], 
                                        overlap=frequency_opts['overlap'], 
                                        step=frequency_opts['step'],
                                        detrend=frequency_opts['detrend'], 
                                        n_fft=n_fft, 
                                        log_trans=False, 
                                        centend=frequency_opts['centend'])
                        
                        freqs = Fooofxx.axis['freq'][0]
                              
                        fooof_powers = zeros((len(bands)))
                        fooof_ap_params = zeros((len(bands), 2))
                        fooof_pk_params = ones((len(bands), 9)) * nan
                        

                        fm.fit(freqs, Fooofxx.data[0][0], fooof_opts['freq_range'])
                        
                        for j, band in enumerate(bands):
                            fp = get_band_peak_fm(fm, band, fooof_opts['select_highest'],
                                                  threshold=fooof_opts['thresh_select'],
                                                  thresh_param=fooof_opts['thresh_param'])
                            if fp.ndim == 1:
                                fooof_powers[j] = gaussian_integral(fp[1], fp[2])
                            else:
                                pwr = asarray([gaussian_integral(fp[x, 1], fp[x, 2]) \
                                               for x in range(fp.shape[0])]).sum()
                                fooof_powers[j] = pwr
                                
                            # get fooof aperiodic parameters
                            fooof_ap_params[j, :] = fm.aperiodic_params_
                            
                            # get fooof peak parameters
                            fp = get_band_peak_fm(fm, band, False,
                                                  threshold=fooof_opts['thresh_select'],
                                                  thresh_param=fooof_opts['thresh_param'])
                            if fp.ndim == 1:
                                fooof_pk_params[j, :3] = fp
                            else:
                                n_peaks = min(fp.shape[0], 3)
                                fooof_pk_params[j, :n_peaks * 3] = fp[:n_peaks, 
                                                                      :].ravel()
                        out['fooof_powers'] = fooof_powers
                        out['fooof_ap_params'] = fooof_ap_params
                        out['fooof_pk_params'] = fooof_pk_params
                                
                        
    
                        seg_info = ['Start time', 'End time', 'Duration', 'Stitches', 
                                    'Stage', 'Cycle', 'Event type', 'Channel']
                        band_hdr = [f'{b1}-{b2} Hz' for b1, b2 in bands]
                        pk_params_hdr = ['peak1_CF', 'peak1_PW', 'peak1_BW', 
                                         'peak2_CF', 'peak2_PW', 'peak2_BW', 
                                         'peak3_CF', 'peak3_PW', 'peak3_BW', ]
                        ap_params_hdr = ['Offset', 'Exponent']
                        band_pk_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                              for p in pk_params_hdr]
                        band_ap_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                              for p in ap_params_hdr]
                        one_record = zeros((1, 
                                            (len(seg_info) + len(bands) + 
                                            len(band_pk_params_hdr) + 
                                            len(band_ap_params_hdr))),
                                            dtype='O')
                        one_record[0, :] = concatenate((asarray([
                                                                out['start'],
                                                                out['end'],
                                                                out['duration'], 
                                                                out['n_stitch'], # number of concatenated segments minus 1
                                                                out['stage'],
                                                                out['cycle'],
                                                                out['name'], # event type
                                                                ch,
                                                                ]),
                                                                out['fooof_powers'],
                                                                out['fooof_pk_params'].ravel(),
                                                                out['fooof_ap_params'].ravel(),
                                                                ))
                        
                        ### WHOLE NIGHT ###
                        if model == 'whole_night':
                            stagename = '-'.join(self.stage)
                            outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_specparams_{suffix}.csv'
                        elif model == 'stage*cycle':    
                            outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}_cycle{cycle_idx[sg]}_specparams_{suffix}.csv'
                        elif model == 'per_stage':
                            outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{self.stage[sg]}_specparams_{suffix}.csv'
                        elif model == 'per_cycle':
                            stagename = '-'.join(self.stage)
                            outputfile = f'{outpath}/{sub}_{ses}_{fnamechan}_{stagename}_cycle{cycle_idx[sg]}_specparams_{suffix}.csv'

                        logger.debug(f'Saving {outputfile}')
                        df = DataFrame(data = one_record, 
                                       columns = (seg_info + band_hdr + band_pk_params_hdr
                                                  + band_ap_params_hdr))
                        df.to_csv(outputfile)
        
        return

    def powerspec_it(self, cat, stage, chan, ref_chan, 
                     general_opts, frequency_opts, rater = None, cycle_idx = None, 
                     subs = 'all', sessions = 'all', filter_opts = None, 
                     epoch_opts = None, event_opts = None, norm = None, 
                     norm_opts = None, fooof_it = False,
                     fooof_opts = None):
        
        print(r"""    Calculating power spectrum...
              
                  |
                  | /\ 
                  |/  \
              uV2 |    ^-_
                  |       ^-___-__
                  |________________
                         (Hz)
              
              """)
            
        if path.exists(self.out_dir):
                print(self.out_dir + " already exists")
        else:
            mkdir(self.out_dir)
        
        
        suffix = general_opts['suffix']
            
        if fooof_it:
            if fooof_opts is None:
                print('Error: Options not set for FOOOF')
            else:
                # Define model parameters
                fm = FOOOF(fooof_opts['peak_width_limits'], fooof_opts['max_n_peaks'], 
                           fooof_opts['min_peak_amplitude'], fooof_opts['peak_threshold'])
            
                def gaussian_integral(a, c):
                    """ Returns definite integral of a gaussian function with height a and
                    standard deviation c."""
                    return sqrt(2) * a * abs(c) * sqrt(pi)
            
            # Prepare Bands
            if fooof_opts['bands_fooof'] is not None:
                bands = fooof_opts['bands_fooof']
            else:
                stp = min(fooof_opts['peak_width_limits'])
                low = int(floor(fooof_opts['freq_range'][0]))
                hi = int(ceil(fooof_opts['freq_range'][1]))
                bands = [(x,x + stp) for x in range(low,hi)]
            
        
        
        ### loop through records
        if isinstance(subs, list):
            None
        elif subs == 'all':
                subs = listdir(self.rec_dir)
                subs = [ p for p in subs if not '.' in p]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' ")  
    
        if norm:
            if norm not in ('integral', 'baseline'):
                exit('Invalid value for norm: ' + str(norm))
                
        for i, p in enumerate(subs):
            # loop through visits
            if sessions == 'all':
                sessions = listdir(self.rec_dir + '/' + p)
                sessions = [x for x in sessions if not '.' in x]    
            
            for v, ses in enumerate(sessions):
                ## Define files
                rdir = self.rec_dir + p + '/' + ses + '/'
                xdir = self.xml_dir + p + '/' + ses + '/'
                edf_file = [x for x in listdir(rdir) if x.endswith('.edf') or x.endswith('.rec') 
                            or x.endswith('.eeg') or x.endswith('.set')]
                xml_file = [x for x in listdir(xdir) if x.endswith('.xml')]            
                
                ### Define output path
                if not path.exists(self.out_dir):
                    mkdir(self.out_dir)
                if not path.exists(self.out_dir + p ):
                    mkdir(self.out_dir + p)
                if not path.exists(self.out_dir + p + '/' + ses):
                    mkdir(self.out_dir + p + '/' + ses)
                outpath = self.out_dir + p + '/' + ses + '/'
                
                
                ## Now import data
                dset = Dataset(rdir + edf_file[0])
                annot = Annotations(xdir + xml_file[0], rater_name=rater)
                 
                ### get cycles
                if cycle_idx is not None:
                    all_cycles = annot.get_cycles()
                    cycle = [all_cycles[i - 1] for i in cycle_idx if i <= len(all_cycles)]
                else:
                    cycle = None
                    
                ### if event channel only, specify event channels
                chan_full = None
                if event_opts['event_chan']:
                    chan_full = [i + ' (' + general_opts['chan_grp_name'] + ')' for i in event_opts['event_chan']]
                
                ### select and read data
                print('Reading data for ' + p + ', Visit: ' + ses)
                segments = fetch(dset, annot, cat=cat, evt_type=event_opts['evt_type'], 
                                 stage=stage, cycle=cycle, chan_full=chan_full, 
                                 epoch=epoch_opts['epoch'], 
                                 epoch_dur=epoch_opts['epoch_dur'], 
                                 epoch_overlap=epoch_opts['epoch_overlap'], 
                                 epoch_step=epoch_opts['epoch_step'], 
                                 reject_epoch=general_opts['reject_epoch'], 
                                 reject_artf=general_opts['reject_artf'],
                                 min_dur=general_opts['min_dur'])
                
                if not segments:
                    print('No valid data found. Skipping to next record.')
                    continue
                
                # BASELINE NORMALISATION
                if norm == 'baseline':
                    norm_seg = fetch(dset, annot, cat=norm_opts['norm_cat'], 
                                     evt_type=norm_opts['norm_evt_type'],
                                     stage=norm_opts['norm_stage'], 
                                     epoch=norm_opts['norm_epoch'])
                    
                    if not norm_seg:
                        print('No valid normalization data found. '
                              'Skipping to next record.')
                        continue
                    
                    if filter_opts['laplacian']:
                        norm_seg.read_data(filter_opts['lapchan'], ref_chan) 
                    else:
                        norm_seg.read_data(chan, ref_chan)            
                    all_nSxx = []
                    
                    for seg in norm_seg:
                        
                        normdata = seg['data']
                        
                        if filter_opts['laplacian']:
                            normdata.data[0] = laplacian_mne(normdata, 
                                                     filter_opts['oREF'], 
                                                     channel=chan, 
                                                     ref_chan=ref_chan, 
                                                     laplacian_rename=filter_opts['laplacian_rename'], 
                                                     renames=filter_opts['renames'])
                        
                        Sxx = frequency(normdata, output=frequency_opts['output'], 
                                        scaling=frequency_opts['scaling'],
                                        sides=frequency_opts['sides'], 
                                        taper=frequency_opts['taper'],
                                        halfbandwidth=frequency_opts['halfbandwidth'], 
                                        NW=frequency_opts['NW'],
                                        duration=frequency_opts['duration'], 
                                        overlap=frequency_opts['overlap'], 
                                        step=frequency_opts['step'],
                                        detrend=frequency_opts['detrend'], 
                                        n_fft=frequency_opts['n_fft'], 
                                        log_trans=frequency_opts['log_trans'], 
                                        centend=frequency_opts['centend'])
                        all_nSxx.append(Sxx)
                        
                        nSxx = ChanFreq()
                        nSxx.s_freq = Sxx.s_freq
                        nSxx.axis['freq'] = Sxx.axis['freq']
                        nSxx.axis['chan'] = Sxx.axis['chan']
                        nSxx.data = empty(1, dtype='O')
                        nSxx.data[0] = empty((Sxx.number_of('chan')[0],
                                 Sxx.number_of('freq')[0]), dtype='f')
                        nSxx.data[0] = mean(
                                stack([x()[0] for x in all_nSxx], axis=2), axis=2)
                
                
                if filter_opts['laplacian']:
                    segments.read_data(filter_opts['lapchan'], ref_chan)
                else:
                    segments.read_data(chan, ref_chan)
                xfreq = []
                
                for sg, seg in enumerate(segments):
                    print(f'Analysing segment {sg} of {len(segments)}')
                    out = dict(seg)
                    data = seg['data']
                    timeline = data.axis['time'][0]
                    out['start'] = timeline[0]
                    out['end'] = timeline[-1]
                    out['duration'] = len(timeline) / data.s_freq
                    
                    if frequency_opts['fast_fft']:
                        n_fft = next_fast_len(data.number_of('time')[0])
                    else:
                        n_fft = frequency_opts['n_fft']
                    
                    if filter_opts['laplacian']:
                        selectchans = filter_opts['lapchan']
                    else:
                        selectchans = chan
                    
                    if filter_opts['notch']:
                        print('Applying notch filtering.')
                        data.data[0] = notch_mne(data, oREF=filter_opts['oREF'], 
                                                    channel=selectchans, 
                                                    freq=filter_opts['notch_freq'],
                                                    rename=filter_opts['laplacian_rename'],
                                                    renames=filter_opts['renames'],
                                                    montage=filter_opts['montage'])
                        
                    if filter_opts['notch_harmonics']: 
                        print('Applying notch harmonics filtering.')
                        print(f'{selectchans}')
                        data.data[0] = notch_mne2(data, oREF=filter_opts['oREF'], 
                                                  channel=selectchans, 
                                                  rename=filter_opts['laplacian_rename'],
                                                  renames=filter_opts['renames'],
                                                  montage=filter_opts['montage'])    
                    
                    if filter_opts['bandpass']:
                        print('Applying bandpass filtering.')
                        data.data[0] = bandpass_mne(data, oREF=filter_opts['oREF'], 
                                                  channel=selectchans,
                                                  highpass=filter_opts['highpass'], 
                                                  lowpass=filter_opts['lowpass'], 
                                                  rename=filter_opts['laplacian_rename'],
                                                  renames=filter_opts['renames'],
                                                  montage=filter_opts['montage'])
                    
                    if filter_opts['laplacian']:
                        print('Applying Laplacian filtering.')
                        data.data[0] = laplacian_mne(data, 
                                             filter_opts['oREF'], 
                                             channel=chan, 
                                             ref_chan=ref_chan, 
                                             laplacian_rename=filter_opts['laplacian_rename'], 
                                             renames=filter_opts['renames'],
                                             montage=filter_opts['montage'])
                        data.axis['chan'][0] = asarray([x for x in chan])
                        selectchans = chan
                    
                    
                    ### frequency transformation
                    Sxx = frequency(data, output=frequency_opts['output'], 
                                    scaling=frequency_opts['scaling'], 
                                    sides=frequency_opts['sides'], 
                                    taper=frequency_opts['taper'],
                                    halfbandwidth=frequency_opts['halfbandwidth'], 
                                    NW=frequency_opts['NW'],
                                    duration=frequency_opts['duration'], 
                                    overlap=frequency_opts['overlap'], 
                                    step=frequency_opts['step'],
                                    detrend=frequency_opts['detrend'], 
                                    n_fft=n_fft, 
                                    log_trans=frequency_opts['log_trans'], 
                                    centend=frequency_opts['centend'])
                    
                    
                    
                    if norm:
            
                        for j, ch in enumerate(Sxx.axis['chan'][0]):
            
                            dat = Sxx.data[0][j,:]
                            sf = Sxx.axis['freq'][0]
                            f_res = sf[1] - sf[0] # frequency resolution
            
                            if norm == 'integral':
                                norm_dat = sum(dat) * f_res # integral by midpoint rule
                            else:
                                norm_dat = nSxx(chan=ch)[0]
            
                            Sxx.data[0][j,:] = dat / norm_dat
            
                    out['data'] = Sxx
                    
                    
                    if fooof_it:
                        
                        Fooofxx = frequency(data, output=frequency_opts['output'], 
                                        scaling=frequency_opts['scaling'], 
                                        sides=frequency_opts['sides'], 
                                        taper=frequency_opts['taper'],
                                        halfbandwidth=frequency_opts['halfbandwidth'], 
                                        NW=frequency_opts['NW'],
                                        duration=frequency_opts['duration'], 
                                        overlap=frequency_opts['overlap'], 
                                        step=frequency_opts['step'],
                                        detrend=frequency_opts['detrend'], 
                                        n_fft=n_fft, 
                                        log_trans=False, 
                                        centend=frequency_opts['centend'])
                        
                        freqs = Fooofxx.axis['freq'][0]
                        
    
                                
                        fooof_powers = zeros((len(chan), len(bands)))
                        fooof_ap_params = zeros((len(chan), len(bands), 2))
                        fooof_pk_params = ones((len(chan), len(bands), 9)) * nan
                        
                        for i in range(len(chan)):
                            fm.fit(freqs, Fooofxx.data[0][i], fooof_opts['freq_range'])
                            
                            for j, band in enumerate(bands):
                                fp = get_band_peak_fm(fm, band, fooof_opts['select_highest'],
                                                      threshold=fooof_opts['thresh_select'],
                                                      thresh_param=fooof_opts['thresh_param'])
                                if fp.ndim == 1:
                                    fooof_powers[i, j] = gaussian_integral(fp[1], fp[2])
                                else:
                                    pwr = asarray([gaussian_integral(fp[x, 1], fp[x, 2]) \
                                                   for x in range(fp.shape[0])]).sum()
                                    fooof_powers[i, j] = pwr
                                    
                                # get fooof aperiodic parameters
                                fooof_ap_params[i, j, :] = fm.aperiodic_params_
                                
                                # get fooof peak parameters
                                fp = get_band_peak_fm(fm, band, False,
                                                      threshold=fooof_opts['thresh_select'],
                                                      thresh_param=fooof_opts['thresh_param'])
                                if fp.ndim == 1:
                                    fooof_pk_params[i, j, :3] = fp
                                else:
                                    n_peaks = min(fp.shape[0], 3)
                                    fooof_pk_params[i, j, :n_peaks * 3] = fp[:n_peaks, 
                                                                          :].ravel()
                        
                        out['fooof_powers'] = fooof_powers
                        out['fooof_ap_params'] = fooof_ap_params
                        out['fooof_pk_params'] = fooof_pk_params
                                
                    xfreq.append(out)
                    
                if general_opts['freq_full']:                        
                    if len(xfreq) == 1:
                        desc = None
                    else:
                        as_matrix = asarray([y for x in xfreq for y in x['data']()[0]])
                        desc = get_descriptives(as_matrix)
                    
                    filename = outpath + p + '_' + ses + f'_freq_full{suffix}.csv'
                    print('Writing to ' + filename)  
                    export_freq(xfreq, filename, desc=desc)
                    
                if general_opts['freq_band']:
                    filename = outpath + p + '_' + ses + f'_freq_band{suffix}.csv'
                    print('Writing to ' + filename)  
                    export_freq_band(xfreq, frequency_opts['bands'], filename)
                    
                if general_opts['freq_plot']:
                    fig = Figure()
                    FigureCanvas(fig)
                    idx_plot = 1
                    
                    for seg in xfreq:
                        data = seg['data']
                        seg_chan = data.axis['chan'][0]
                        sf = data.axis['freq'][0]
                        if general_opts['max_freq_plot']:
                            idx_max = asarray(
                                    [abs(x - general_opts['max_freq_plot']) for x in sf]).argmin()
                        for ch in seg_chan:
                            Sxx = data(chan=ch)[0]
                            ax = fig.add_subplot(len(xfreq), len(seg_chan), idx_plot)
                            ax.semilogy(sf[1:idx_max], Sxx[1:idx_max])
                            ax.set_xlabel('Frequency (Hz)')
                            
                            idx_plot += 1
                        
                    print('Saving figure to ' + outpath + p + '.png')
                    fig.savefig(outpath + p + suffix + '.png')
                    
                if fooof_it:
                    seg_info = ['Start time', 'End time', 'Duration', 'Stitches', 
                        'Stage', 'Cycle', 'Event type', 'Channel']
                    #band_hdr = [str(b1) + '-' + str(b2) for b1, b2 in bands_fooof]
                    band_hdr = [f'{b1}-{b2} Hz' for b1, b2 in bands]
                    pk_params_hdr = ['peak1_CF', 'peak1_PW', 'peak1_BW', 
                                  'peak2_CF', 'peak2_PW', 'peak2_BW', 
                                  'peak3_CF', 'peak3_PW', 'peak3_BW', ]
                    ap_params_hdr = ['Offset', 'Exponent']
                    band_pk_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                          for p in pk_params_hdr]
                    band_ap_params_hdr = ['_'.join((b, p)) for b in band_hdr 
                                          for p in ap_params_hdr]
                    one_record = zeros((len(xfreq) * len(chan), 
                                        (len(seg_info) + len(bands) + 
                                        len(band_pk_params_hdr) + 
                                        len(band_ap_params_hdr))),
                                        dtype='O')
                    for i, seg in enumerate(xfreq):
                        for j, ch in enumerate(chan):
                            cyc = None
                            if seg['cycle'] is not None:
                                cyc = seg['cycle'][2]
                            
                            one_record[i * len(chan) + j, :] = concatenate((asarray([
                                seg['start'],
                                seg['end'],
                                seg['duration'], 
                                seg['n_stitch'], # number of concatenated segments minus 1
                                seg['stage'],
                                cyc,
                                seg['name'], # event type
                                ch,
                                ]),
                                seg['fooof_powers'][j, :],
                                seg['fooof_pk_params'][j, ...].ravel(),
                                seg['fooof_ap_params'][j, ...].ravel(),
                                ))
                            
                    outpath_fooof = outpath + p + '_' + ses + f'_fooofbands{suffix}.csv'
                    print(f'Saving {outpath_fooof}')
                    df = DataFrame(data=one_record, 
                                   columns=(seg_info + band_hdr + band_pk_params_hdr
                                            + band_ap_params_hdr))
                    df.to_csv(outpath_fooof)
        
        print("Job's done.")
        
        
        
    def powerspec_summary_bands(root_dir, out_dir, part, visit, chan, stage, col_headers, 
                          excel_output_file):
    
        ## Script will create a file with multiple tab: once per STAGE*CHANNEL
        
        """ SCRIPT """
        csvs = []
        which_files = '*_freq_band.csv' # Name extension of output files to search for
        tabs = [' '.join((c,s)) for c in chan for s in stage] # Create matrices for each stage*channel combo
        
        # Loop through participants and visits to extract individual CSV output files
        if isinstance(part, list):
                None
        elif part == 'all':
                part = listdir(root_dir)
                part = [x for x in part if not('.' in x)]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' ")      
                
        for i, p in enumerate(part):
                print(p)
                # loop through visits
                if visit == 'all':
                    visit = listdir(root_dir + '/' + p)
                    visit = [x for x in visit if not('.' in x)]
                
                for v, vis in enumerate(visit):
                    ## Define files
                    rdir = root_dir + p + '/' + vis + '/'
                    
                    try:
                        csvs.append(glob(f'{rdir}/{which_files}')[0]) #glob= va chercher tous les path du filename
                    except Exception as e:
                        print(f'No files found for {p}, visit {vis}')
                    
        ids = [x + '_' + y for x in part for y in visit] 
        idx_data_col = list(range(9,9+len(col_headers)))
        data = ones((len(chan) * len(stage), len(ids), len(idx_data_col))) * nan #create matrice - *nan to make sure that missing value will be blank
            
        for i, one_csv in enumerate(csvs): #on va passer dans la liste 1 a 1 (commence a compter a 0)    
            with open(one_csv, newline='') as f: #with open = to open a text/csv file
                csv_reader = reader(f, delimiter=',') #f = file, delimiter = what separate value in xcl (;, ou, ou .)
                
                for row in range(2): #skip first line
                    row = next(csv_reader) #next = move forward
                    
                if col_headers is None:
                    col_headers = [row[x] for x in idx_data_col] #if None in col_header then will use the name of the csv
                
                for row in csv_reader:
                    try: #try below, but if valueerror continue to except
                        int(row[0]) #find integer (chiffre)
                        idx_tab = tabs.index(' '.join((row[8], row[5]))) #join column 8 (channel) and 5(stage) to know which tab to go to
                        
                        for j, idx_col in enumerate(idx_data_col):
                            data[idx_tab, i, j] = row[idx_col] #create data, 1 column at a time and write it in correct tab
                                        
                    except ValueError:
                        continue
        
        wb = Workbook()
        wb.save(excel_output_file)
                       
        with ExcelWriter(excel_output_file, engine="openpyxl", mode='a') as writer: #create xcl
        
            for i, tab_label in enumerate(tabs):
                df = DataFrame(data[i, ...], index=ids, columns=col_headers) #create dataframe and write data in it
                df.to_excel(writer, sheet_name=tab_label) #method of df to create xcl
            
        print(f'Saved to {excel_output_file}')
        
        
        
    def powerspec_summary_full(root_dir, out_dir, part, visit, chan, stage, lowpass,
                          excel_output_file):
    
        ## Script will create a file with multiple tab: once per STAGE*CHANNEL
        
        """ SCRIPT """
        csvs = []
        which_files = '*_freq_full.csv' # Name extension of output files to search for
        tabs = [' '.join((c,s)) for c in chan for s in stage] # Create matrices for each stage*channel combo
        
        # Loop through participants and visits to extract individual CSV output files
        if isinstance(part, list):
                None
        elif part == 'all':
                part = listdir(root_dir)
                part = [x for x in part if not('.' in x)]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' ")      
                
        for i, p in enumerate(part):
                print(p)
                # loop through visits
                if visit == 'all':
                    visit = listdir(root_dir + '/' + p)
                    visit = [x for x in visit if not('.' in x)]
                
                for v, vis in enumerate(visit):
                    ## Define files
                    rdir = root_dir + p + '/' + vis + '/'
                    
                    try:
                        csvs.append(glob(f'{rdir}/{which_files}')[0]) #glob= va chercher tous les path du filename
                    except Exception as e:
                        print(f'No files found for {p}, visit {vis}')
                    
        ids = [x + '_' + y for x in part for y in visit] 
        idx_data_col = list(range(9,9+lowpass*4))
        data = ones((len(chan) * len(stage), len(ids), len(idx_data_col))) * nan #create matrice - *nan to make sure that missing value will be blank
            
        for i, one_csv in enumerate(csvs): #on va passer dans la liste 1 a 1 (commence a compter a 0)    
            with open(one_csv, newline='') as f: #with open = to open a text/csv file
                csv_reader = reader(f, delimiter=',') #f = file, delimiter = what separate value in xcl (;, ou, ou .)
                
                for row in range(2): #skip first line
                    row = next(csv_reader) #next = move forward
                    
                
                col_headers = [row[x] for x in idx_data_col] #if None in col_header then will use the name of the csv
                
                for row in csv_reader:
                    try: #try below, but if valueerror continue to except
                        int(row[0]) #find integer (chiffre)
                        idx_tab = tabs.index(' '.join((row[8], row[5]))) #join column 8 (channel) and 5(stage) to know which tab to go to
                        
                        for j, idx_col in enumerate(idx_data_col):
                            data[idx_tab, i, j] = row[idx_col] #create data, 1 column at a time and write it in correct tab
                                        
                    except ValueError:
                        continue
        
        wb = Workbook()
        wb.save(excel_output_file)
                       
        with ExcelWriter(excel_output_file, engine="openpyxl", mode='a') as writer: #create xcl
        
            for i, tab_label in enumerate(tabs):
                df = DataFrame(data[i, ...], index=ids, columns=col_headers) #create dataframe and write data in it
                df.to_excel(writer, sheet_name=tab_label) #method of df to create xcl
            
        print(f'Saved to {excel_output_file}')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 15 11:33:16 2022

@author: nathancross
"""
from .logs import create_logger
import mne
import numpy as np
from pandas import DataFrame
import copy
from operator import itemgetter
from os import listdir, path, mkdir
from datetime import datetime
from wonambi import Dataset, graphoelement
from wonambi.attr.annotations import Annotations, create_empty_annotations
from scipy.signal import find_peaks, periodogram
import shutil
from sleepecg import detect_heartbeats



def clean_annots(xml_dir, out_dir, rater, keep_evts = None, subs = 'all', 
                 sessions = 'all', logger = create_logger('Clean annotations')):
    
    '''
        Copies the annotations files from inside xml_dir into out_dir and 
        removes all annotations from the files.
        
    '''
    
    logger.debug('')
    logger.debug(f'Time start: {datetime.now()}')
    logger.debug(f'Removing events from files in directory {xml_dir}')  
    logger.debug(f'Keeping events = {keep_evts}')
    logger.debug(f'Saving new annotations files to directory {out_dir}')
    logger.debug('')
    
    # Loop through subjects
    if isinstance(subs, list):
        None
    elif subs == 'all':
            subs = listdir(xml_dir)
            subs = [x for x in subs if not '.' in x]
    else:
        logger.error("'subs' must either be an array of subject ids or = 'all' **CHECK**")
    
    subs.sort()                
    for i, sub in enumerate(subs):
        # Loop through visits
        if sessions == 'all':
            sessions = listdir(xml_dir + '/' + '/' + sub)
            sessions = [x for x in sessions if not '.' in x]
        sessions.sort()
        for v, ses in enumerate(sessions): 
            
            # Define files
            xdir = xml_dir + '/' + sub + '/' + ses + '/'
            xml_files = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_files) == 0:                
                logger.warning(f'No xml files found for Subject {sub}, skipping..')
  
            else:
                for xml_file in xml_files:
                    ## Copy annotations file before beginning
                    if not path.exists(out_dir):
                        mkdir(out_dir)
                    if not path.exists(out_dir + sub):
                        mkdir(out_dir + sub)
                    if not path.exists(out_dir + sub + '/' + ses):
                        mkdir(out_dir + sub + '/' + ses)
                    backup = out_dir + sub + '/' + ses + '/'
                    backup_file = (f'{backup}{xml_file[0]}')
                    shutil.copy(xdir + xml_file, backup_file)
                                
                    # Import Annotations file
                    annot = Annotations(backup_file, rater_name=rater)
                    evts = list(set([x['name'] for x in annot.get_events()]))
                    
                    logger.debug(f'Removing events for {sub}, {ses}..')
                    if keep_evts == None:
                        keep_evts = []
                    for ev in evts:
                        if ev not in keep_evts:
                            annot.remove_event_type(name=ev)
                

def remove_event(annot, evt_name, chan = None, stage = None):
    
    ''' Remove event from annotation
        Workaround function because wonambi.attr.annotations.remove_event()
        is not working properly (and I couldn't figure out why).
    '''

    if not chan and not stage:
        evts = [x for x in annot.get_events(name=evt_name)]
    elif not stage:
        evts = [x for x in annot.get_events(name=evt_name) 
                if chan not in x['chan']]
    elif not chan:
        evts = [x for x in annot.get_events(name=evt_name) 
                if stage not in x['stage']]
    
    annot.remove_event_type(name=evt_name)
    grapho = graphoelement.Graphoelement()
    grapho.events = evts       
    grapho.to_annot(annot, evt_name)
    
    return
    

def remove_duplicate_evts(annot, evt_name, chan, stage = None):
    
    ''' Removes any events in annotations that are duplicate.
        Workaround function because wonambi.attr.annotations.remove_event()
        is not working properly (and I couldn't figure out why).
    '''
    evts = annot.get_events(name=evt_name, chan = chan, stage = stage)
    evts_trim = copy.deepcopy(evts)
    for e, event in enumerate(evts[:-1]):
        starttime = event['start']
        i = 0
        for ee, eevent in enumerate(evts_trim):
                if eevent['start'] == starttime:
                    if i == 0:
                        None
                    elif i >0:
                        del(evts_trim[ee])
                    i=+1
    
    evts = [x for x in annot.get_events(name=evt_name) if chan not in x['chan']]
    
    annot.remove_event_type(name=evt_name)
    grapho = graphoelement.Graphoelement()
    grapho.events = evts + evts_trim          
    grapho.to_annot(annot, evt_name)
    
    return


def remove_duplicate_evts_bids(in_dir, out_dir, chan, grp_name, rater, 
                               cat=(0, 0, 0, 0), stage = None, 
                               evt_name = None, part = 'all', visit = 'all', 
                               param_keys = None):

    if path.exists(in_dir):
            print(in_dir + " already exists")
    else:
        mkdir(in_dir)
    
    print('')
    print(f'Time start: {datetime.now()}')
    print(f'Removing duplicates from files in directory {in_dir}')  
    print(f'Event = {evt_name[0]}')
    print('')
    
    if evt_name == None:
        evt_name = ['spindle']
    
    # Loop through subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(in_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    part.sort()                
    for i, p in enumerate(part):
        # Loop through visits
        if visit == 'all':
            visit = listdir(in_dir + '/' + p)
            visit = [x for x in visit if not '.' in x]
        
        print(f'Removing duplicate events for Subject {p}..')
        
        visit.sort()
        for v, vis in enumerate(visit): 
            if not path.exists(out_dir + '/' + p + '/' + vis + '/'):
                mkdir((out_dir + '/' + p + '/' + vis + '/'))
            backup = out_dir + '/' + p + '/' + vis + '/' #backup folder

            # Define files
            rec_dir = in_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(rec_dir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_file) == 0:                
                print(f'WARNING: whale_it has not been run for Subject {p}, skipping..')
                
            else:
                # backup file
                backup_file = (backup + xml_file[0])
                shutil.copy(rec_dir + xml_file[0], backup_file) 
                
                # Import Annotations file
                annot = Annotations(backup_file, 
                                    rater_name=rater)
                
                # Run through channels
                for ch, channel in enumerate(chan):
                    chan_ful = channel + ' (' + grp_name + ')'
                                    
                    ### WHOLE NIGHT ###
                    # Select and read data
                    print('Reading data for ' + p + ', visit ' + vis + ' ' + channel)
                    
                    # Retrieve events
                    remove_duplicate_evts(annot, evt_name, chan_ful, stage)

    return


def merge_xmls(in_dirs, out_dir, chan, grp_name, stage = None, evt_name = None, 
               part = 'all', visit = 'all'):

    for pth in in_dirs:
        if not path.exists(pth):
            print(pth + " doesn't exist. Check in_dirs is set correctly.")
            
    else: 
        main_dir = in_dirs[0]
        sec_dir = in_dirs[1]
        
        print('')
        print(f'Time start: {datetime.now()}')
        print(f'Initiating the merging of xml files in directory {main_dir}...') 
        print(f'with the xml files in directory {sec_dir}...') 
        print('')
        
        # Get list of subjects
        if isinstance(part, list):
            None
        elif part == 'all':
                part = listdir(main_dir)
                part = [ p for p in part if not '.' in p]
        else:
            print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
        
        
        # Loop through subjects  
        part.sort()              
        for i, p in enumerate(part):
            print(f'Merging xml files for Subject {p}..')
            # Get visits
            if visit == 'all':
                visit = listdir(main_dir + '/' + p)
                visit = [x for x in visit if not '.' in x]
            
            # Loop through visits
            visit.sort()
            for v, vis in enumerate(visit): 
                if not path.exists(main_dir + '/' + p + '/' + vis):
                    print(main_dir + '/' + p + '/' + vis + " doesn't exist. Check data is in BIDS format.")
                
                # Check for output directory and create
                if not path.exists(out_dir + '/'):
                     mkdir(out_dir + '/')
                if not path.exists(out_dir + '/' + p):
                    mkdir(out_dir + '/' + p)
                if not path.exists(out_dir + '/' + p + '/' + vis):
                   mkdir(out_dir + '/' + p + '/' + vis)
                output = out_dir + '/' + p + '/' + vis + '/' #backup folder
    
    
                # Define files
                if main_dir == sec_dir:
                
                    rec_dir = main_dir + '/' + p + '/' + vis + '/'
                    xml_file = [x for x in listdir(rec_dir) if x.endswith('.xml') 
                                if not x.startswith('.')] 
                    xml_file.sort()
                    
                else:
                    rec_dir = main_dir + '/' + p + '/' + vis + '/'
                    xml_file1 = [x for x in listdir(rec_dir) if x.endswith('.xml') 
                                if not x.startswith('.')]
                    rec_dir2 = sec_dir + '/' + p + '/' + vis + '/'
                    xml_file2 = [x for x in listdir(rec_dir2) if x.endswith('.xml') 
                                if not x.startswith('.')]
                    xml_file = xml_file1 + xml_file2
                    
                    
                # Merge files
                if len(xml_file) < 2:                
                    print(f'WARNING: Only 1 xml file found for Subject {p}, skipping..')
                else:
                    # Copy first file to output directory
                    output_file = (output + p + '-merged.xml')
                    shutil.copy(rec_dir + xml_file[0], output_file) 
                    
                    # Import Annotations file
                    annot = Annotations(output_file, 
                                        rater_name=None)
                    
                    for x,xml in enumerate(xml_file[1:]):
                        annot2 = Annotations(rec_dir + xml, 
                                        rater_name=None)
                    
                        # Run through channels
                        for ch, channel in enumerate(chan):
                            chan_ful = [channel + ' (' + grp_name + ')']
                                            
                            print('Merging on channel: {channel}')
                            
                            # Calculate event density (whole night)
                            if evt_name is not None:
                                for e,evt in enumerate(evt_name):
                                    evts = annot2.get_events(name=evt, chan = chan_ful, 
                                                             stage = stage)
                                    annot.add_events(evts)
                            else:
                                evts = annot2.get_events(name=None, chan = chan_ful, 
                                                         stage = stage)
                                annot.add_events(evts)

    return


def rainbow_merge_evts(xml_dir, out_dir, chan, grp_name, rater, segments = None, 
                       events = None, evt_name = None, part = 'all', 
                       visit = 'all'):
    
    # Make output directory
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    if evt_name == None:
        evt_name = ['spindle']
    
    # Loop through subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not(p.startswith('.'))]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    part.sort()                
    for i, p in enumerate(part):
        # Loop through visits
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not(x.startswith('.'))]
        
        if not path.exists(out_dir + '/' + p):
            mkdir(out_dir + '/' + p)
        print(" ")      
        print(f'Merging rainbows for Subject {p}..')
        print(r""" 
                 .##@@&&&@@##.
              ,##@&::%&&%%::&@##.
             #@&:%%000000000%%:&@#
           #@&:%00'         '00%:&@#
          #@&:%0'             '0%:&@#
         #@&:%0                 0%:&@#
        #@&:%0                   0%:&@#
        #@&:%0                   0%:&@#
                """,flush=True)   
        
        
        visit.sort()
        for v, vis in enumerate(visit): 
            if not path.exists(out_dir + '/' + p + '/' + vis + r'/'):
                mkdir(out_dir + '/' + p + '/' + vis + r'/')
            backup = out_dir + '/' + p + '/' + vis + r'/' #backup folder

            # Define files
            xdir = xml_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            
            if len(xml_file) == 0:                
                print(f'WARNING: no xml files found for Subject {p}, skipping..')
            elif len(xml_file) > 1:
                print(f'WARNING: multiple xml files found for Subject {p}, visit {v}. Check this! Skipping..')
            else:
                # backup file
                backup_file = (backup + p + '_trunc.xml')
                shutil.copy(xdir + xml_file[0], backup_file) 
                
                # Import Annotations file
                annot = Annotations(backup_file,rater_name=rater)
                if segments is not None:
                    times = []
                    for seg in segments:
                        
                        starts = [x['end'] for x in sorted(annot.get_events(name=seg[0]), 
                                                           key=itemgetter('start')) ]
                        ends = [x['start'] for x in sorted(annot.get_events(name=seg[1]), 
                                                           key=itemgetter('start')) ]
                        for i in range(0,len(starts)):
                            times.append((starts[i],ends[i]))
                else:
                    times = [(0,annot.last_second)]
                
                evts = []
                for t,time in enumerate(times):
                    
                    if events is not None:
                        for e in events:
                            evts.extend(annot.get_events(name=e,time=time)[:])
                    else:
                        evts = annot.get_events(time=time)[:]

            
                names = set([e['name'] for e in evts])  
                newevents = copy.deepcopy(evts)
                d1 = {'name': evt_name[0]}
                [x.update(d1) for x in newevents]
                

                for t,typ in enumerate(names):    
                    annot.remove_event_type(name=typ)
                
                grapho = graphoelement.Graphoelement()
                grapho.events = newevents          
                grapho.to_annot(annot)


    return


def rename_evts(xml_dir, out_dir, part, visit, evts):
    
    '''
       Renames events inside annotations files. The xmls are first copied to out_dir
       and the events are deleted from the copied xmls.
       Inputs are :
           xml_dir, out_dir = input xml_directory and output directory for new xmls
           evts             = dictionary for how events are to be renamed:
                               e.g. evts = {'Moelle2011':'spindle',
                                            'Lacourse2018':'spindle,
                                            'Staresina2015':'so'}
    '''
    if not path.exists(out_dir):
        mkdir(out_dir)
    
    # Get list of subjects
    if isinstance(part, list):
        None
    elif part == 'all':
            part = listdir(xml_dir)
            part = [ p for p in part if not '.' in p]
    else:
        print("ERROR: 'part' must either be an array of subject ids or = 'all' **CHECK**")
    
    
    # Loop through subjects  
    part.sort()              
    for i, p in enumerate(part):
        
        # Get visits
        if visit == 'all':
            visit = listdir(xml_dir + '/' + p)
            visit = [x for x in visit if not '.' in x]
        
        # Loop through visits
        visit.sort()
        for v, vis in enumerate(visit): 
            
            print(f'Subject {p}, visit {vis}..')
            
            # Check for input data
            if not path.exists(xml_dir + '/' + p + '/' + vis):
                print(xml_dir + '/' + p + '/' + vis + " doesn't exist. Check data is in BIDS format.")
            
            # Check for output directory and create
            if not path.exists(out_dir + '/'):
                 mkdir(out_dir + '/')
            if not path.exists(out_dir + '/' + p):
                mkdir(out_dir + '/' + p)
            if not path.exists(out_dir + '/' + p + '/' + vis):
               mkdir(out_dir + '/' + p + '/' + vis)
            output = out_dir + '/' + p + '/' + vis + '/' #backup folder

            # Define files
            xdir = xml_dir + '/' + p + '/' + vis + '/'
            xml_file = [x for x in listdir(xdir) if x.endswith('.xml') if not x.startswith('.')] 
            xml_file.sort()
            
            if len(xml_file) > 1:                
                print(f'WARNING: Multiple xml files for Subject {p}, visit {vis} - check this - skipping..')
                
            else:
                
                # Copy first file to output directory
                output_file = (output + xml_file[0])
                shutil.copy(xdir + xml_file[0], output_file) 
                
                # Import Annotations file
                annot = Annotations(output_file, rater_name=None)
                
                # Rename events
                for e,event in enumerate(evts):
                    old = event
                    new = evts[event]
                    
                    print(f"Renaming event '{old}' to '{new}'")
                    
                    annot.rename_event_type(old, new)
            
                
    return
                

def laplacian_mne(data, oREF, channel, ref_chan, laplacian_rename=False, 
                  renames=None, montage='standard_alphabetic'):
    
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if oREF:
        dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
    
    if laplacian_rename:
        dig.rename_channels(renames, allow_duplicates=False)
    
    info = mne.create_info(ch_names, data.s_freq,verbose=None)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    a.set_eeg_reference(ref_chan,verbose=40)
    raw_csd = mne.preprocessing.compute_current_source_density(a, verbose=40)
    data = raw_csd.get_data(picks=channel)
    
    return data



def notch_mne(data, oREF, channel, freq, rename=False,
                 renames=None, montage='standard_alphabetic'):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
    
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    anotch = a.notch_filter(freq,verbose=40)
    data = anotch.get_data(picks=channel)
    
    return data

def notch_mne2(data, oREF, channel, rename=False, renames=None,
               montage='standard_alphabetic'):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    anotch = a.notch_filter(freqs=None, filter_length='auto', 
                            notch_widths=None, trans_bandwidth=1, 
                            method='spectrum_fit', verbose=None)
    data = anotch.get_data(picks=channel)
    
    return data
    

def bandpass_mne(data, oREF, channel, highpass, lowpass, rename=False,
                 renames=None, montage='standard_alphabetic'):
    
    ch_names = list(data.axis['chan'][0])
    dig = mne.channels.make_standard_montage(montage)
    
    if rename:
        dig.rename_channels(renames, allow_duplicates=False)
        
    if oREF:
        dig.rename_channels({oREF:'_REF'}, allow_duplicates=False)
        
    info = mne.create_info(ch_names, data.s_freq,verbose=40)
    mneobj = mne.io.RawArray(data.data[0],info,verbose=40)
    dic = [{x:'eeg' for x in mneobj.ch_names}]
    mneobj.set_channel_types(dic[0],verbose=40)
    mneobj.info.set_montage(dig,verbose=40)
    a = mneobj.pick(['eeg']).load_data()
    a_filt = a.filter(highpass, lowpass,verbose=40)
    data = a_filt.get_data(picks=channel)
    
    return data

def csv_stage_import(edf_file, xml_file, hypno_file, rater):
    
    ''' This function creates a new annoations file and imports staging from a csv 
        file. The staging should be in the 'Alice' format, for further information 
        see: wonambi.attr.annotations.
    '''

    dataset = Dataset(edf_file)
    create_empty_annotations(xml_file, dataset)
    annot = Annotations(xml_file)
    annot.import_staging(hypno_file, 'alice', rater_name=rater, rec_start=dataset.header['start_time'])
    
    
    
def infer_eog(dset, logger = create_logger('Infer EOG')):
    
    ''' This function reads a recording file (e.g. edf) and searches for likely
        EOG channel names. IF the EOG channels are not easly identifiable from
        the edf, then nothing is returned. IF there are multiple (>2) options,
        then the channels will be QC'd and the channels with the best quality 
        will be returned.
    '''
    
    h = dset.header
    
    # First and most obvious EOG name
    eyes = [x for x in h['chan_name'] if 'EOG' in x]  
    
    # If EOG are not labelled as 'EOG'
    if len(eyes) == 0:
        eyes = [x for x in h['chan_name'] if 'E' in x if not 'EMG' in x 
                                                      if not 'ECG' in x]    
    
    # If still nothing or when all electrode names contain 'E' (e.g. EGI nets) 
    if len(eyes) == 0 or len(eyes) > 6:
        logger.error('Unable to determine EOG channels from recording. EOG names must be explicitly specified.')
        logger.info('Check documentation for how to specify channels:')
        logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
        logger.info('-' * 10)
        return None

    elif len(eyes) > 2:
        
        loc = [x for x in eyes if any(l in x for l in ['l', 'L'])]
        loc = choose_best_eog(loc, dset, 1, logger)
        
        roc = [x for x in eyes if any(l in x for l in ['r', 'R'])]
        roc = choose_best_eog(roc, dset, 1, logger)
        
        eyes = roc + loc
            
    return eyes    
        
### Quality Control Checks    
def gini(x, w = None):
    
    ''' Calculates the Gini coefficient:
                    a measure of statistical dispersion calculated by comparing 
                    the actual Lorenz curve to the diagonal line of equality.
    '''

    x = np.asarray(x)
    if w is not None:
        w = np.asarray(w)
        sorted_indices = np.argsort(x)
        sorted_x = x[sorted_indices]
        sorted_w = w[sorted_indices]
        # Force float dtype to avoid overflows
        cumw = np.cumsum(sorted_w, dtype=float)
        cumxw = np.cumsum(sorted_x * sorted_w, dtype=float)
        return (np.sum(cumxw[1:] * cumw[:-1] - cumxw[:-1] * cumw[1:]) / 
                (cumxw[-1] * cumw[-1]))
    else:
        sorted_x = np.sort(x)
        n = len(x)
        cumx = np.cumsum(sorted_x, dtype=float)
        # The above formula, with all weights equal to 1 simplifies to:
        return (n + 1 - 2 * np.sum(cumx) / cumx[-1]) / n

def moving_average_time(a, s_freq, win=5):
    n = win*s_freq*60
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

def moving_average(a, win=100):
    n = win
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


def choose_best_eeg(chans, dset, num = 2, logger = create_logger('QC EEG')):
    
    quals = DataFrame(data=None, dtype = float, 
                          columns=['integral',
                                  'stdev',
                                  'time_not_in_low_range',
                                  'time_not_flatline',
                                  '99th_quantile',
                                  'gini_coeff'])
    
    ch_dat = dset.read_data(chan = chans)
    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        e = moving_average_time(abs(ch_dat.data[0][i]),s_freq,5)
        quals.loc[i,'integral'] = sum(e) # (AUC)
        quals.loc[i,'stdev'] = np.std(e)
        quals.loc[i,'time_not_in_low_range'] = 1 - sum(np.logical_and(e>=0, e<=0.5))/len(e)
        quals.loc[i,'time_not_flatline'] = 1 - (sum(np.diff(e)==0)/len(np.diff(e)))
        quals.loc[i,'99th_quantile'] = np.quantile(e,0.99)
        quals.loc[i,'gini_coeff'] = gini(e)
    
    winners = DataFrame([quals[x].rank() for x in quals.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    eeg_names = [chans[x] for x in idx]
    logger.debug(f'Best EEG channels based on auto-QC are: {eeg_names}')
    
    return eeg_names



def choose_best_eog(chans, dset, num = 2, logger = create_logger('QC EOG')):
    
    quals = DataFrame(data=None, dtype = float, 
                          columns=['integral',
                                  'stdev',
                                  'time_not_in_low_range',
                                  'time_not_flatline',
                                  '99th_quantile',
                                  'gini_coeff'])
    
    ch_dat = dset.read_data(chan = chans)
    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        e = moving_average_time(abs(ch_dat.data[0][i]),s_freq,5)
        quals.loc[i,'integral'] = sum(e) # (AUC)
        quals.loc[i,'stdev'] = np.std(e)
        quals.loc[i,'time_not_in_low_range'] = 1 - sum(np.logical_and(e>=0, e<=0.5))/len(e)
        quals.loc[i,'time_not_flatline'] = 1 - (sum(np.diff(e)==0)/len(np.diff(e)))
        quals.loc[i,'99th_quantile'] = np.quantile(e,0.99)
        quals.loc[i,'gini_coeff'] = gini(e)
    
    winners = DataFrame([quals[x].rank() for x in quals.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    eog_names = [chans[x] for x in idx]
    
    
    
    logger.debug(f'Best EOG channels based on auto-QC are: {eog_names}')
    
    return eog_names




def choose_best_ecg(chans, dset, num = 1, logger = create_logger('QC ECG')):
    
    quals = DataFrame(data=None, columns=['hr_qual',
                                          'stdev',
                                          'time_not_flatline',
                                          'PSD_qual'])
    ecg_dat = dset.read_data(chan=chans)
    
    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        e = ecg_dat.data[0][i]
        
        # Check BPM
        beats = detect_heartbeats(e, s_freq)
        hr = len(beats)/(len(e)/(s_freq*60))
        if np.logical_and(hr>40, hr<120):
            logger.debug(f'Heart rate ({round(hr)}bpm) seems physiologically plausible. Continuing...')
            quals.loc[i,'hr_qual'] = 2
        elif np.logical_and(hr>120, hr<200):
            quals.loc[i,'hr_qual'] = 1
            logger.warning(f'Heart rate ({round(hr)}bpm) is physiologically plausible but high. Check...')
        else:
            quals.loc[i,'hr_qual'] = 0
            logger.warning(f'Heart rate ({round(hr)}bpm) is NOT physiologically plausible. Signal quality is likely poor...')
        
        # Check Std Dev.
        quals.loc[i,'stdev'] = np.std(abs(e)) #standard deviation 
        
        # Check for flatlines
        quals.loc[i,'time_not_flatline'] = 1 - (sum(np.diff(e)==0)/len(np.diff(e)))
        
        # Check PSD        
        f_p, Pxx_spec = periodogram(e, s_freq)
        Pxx_spec_sm = np.concatenate((np.zeros(2000),
                                      moving_average(Pxx_spec, 4000),
                                      np.zeros(1999)))
        top = int((len(e)/s_freq)*20) # corresponds to 20Hz
        peaks = find_peaks(Pxx_spec_sm[:top], prominence=800)[0]
        
        if len(peaks) > 4:
            quals.loc[i,'PSD_qual'] = 2
        elif len(peaks) > 2:
            quals.loc[i,'PSD_qual'] = 1
        else:
            print(f'WARNING: PSD of {name} looks unusual..')
            quals.loc[i,'PSD_qual'] = 0

    winners = DataFrame([quals[x].rank() for x in quals.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    ecg_name = [chans[x] for x in idx]
    logger.debug(f'Best ECG channel based on auto-QC is: {ecg_name}')

    return ecg_name


def choose_best_emg(chans, dset, num = 1, logger = create_logger('QC EMG')):

    emg_dat = dset.read_data(chan=chans)
    quals = DataFrame(data=None, dtype = float, 
                      columns=['integral',
                               'stdev',
                               'time_not_in_low_range',
                               'time_not_flatline',
                               '99th_quantile',
                               'gini_coeff'])
    
    for i, name in enumerate(chans):
        logger.debug(f'Checking quality of {name}')
        s_freq = dset.header['orig']['n_samples_per_record'][chans.index(name)]
        e = moving_average_time(abs(emg_dat.data[0][i]),s_freq,5)
        quals.loc[i,'integral'] = sum(e) #intragral (AUC)
        quals.loc[i,'stdev'] = np.std(e) #standard deviation 
        quals.loc[i,'time_not_in_low_range'] = 1 - sum(np.logical_and(e>=0, e<=0.5))/len(e)
        quals.loc[i,'time_not_flatline'] = 1 - (sum(np.diff(e)==0)/len(np.diff(e)))
        quals.loc[i,'99th_quantile'] = np.quantile(e,0.99)
        quals.loc[i,'gini_coeff'] = gini(e)
    
    winners = DataFrame([quals[x].rank() for x in quals.columns]).T
    idx = winners.sum(axis = 1).nlargest(num).index.to_list()
    emg_name = [chans[x] for x in idx]
    print(f'Best EMG channel based on auto-QC is: {emg_name}')
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jul 25 12:07:36 2023

@author: nathancross
"""
from datetime import datetime, date
from os import listdir, mkdir, path, remove, walk
from pandas import DataFrame, read_csv
from seapipe.events.fish import FISH
from seapipe.events.whales import whales
from seapipe.events.remora import remora
from seapipe.events.seasnakes import seasnakes
from seapipe.events.seabass import seabass
from seapipe.events.sand import SAND
from seapipe.pac.octopus import octopus, pac_method
from seapipe.pac.pacats import pacats
from seapipe.spectrum.psa import (Spectrum, default_epoch_opts, default_event_opts,
                     default_fooof_opts, default_filter_opts, default_frequency_opts, 
                     default_general_opts,default_norm_opts)
from seapipe.spectrum.spectrogram import event_spectrogram, event_spectrogram_grouplevel
from seapipe.stats import sleepstats
from seapipe.utils.audit import (check_dataset, check_fooof, extract_channels, make_bids,
                        track_processing)
from seapipe.utils.logs import create_logger, create_logger_outfile
from seapipe.utils.load import (check_chans, check_adap_bands, read_tracking_sheet, 
                                select_input_dirs, select_output_dirs, load_stages)


## TO DO:
#   - adapt load channels to be flexible for non-equivalent refsets and chansets
#   - add in log for detection whether auto, fixed or adapted bands was run
#   - add logging to save to output file (not implemented for all functions)
#   - update adapted bands in tracking.tsv
#  ** create catch for errors in tracking sheet around ',' (chans, adap_bands etc.)
#   - fix discrepency between track WARNINGS and output in dataframe 
#   - update initial tracking to include spindles, slow_oscillation, cfc, power_spectrum
#   - update export sleepstats to export by stage/cycle separately
#   - possibility of cycle_idx = 'all'
#   - enable macro_dataset per sleep cycle
#   - enable downsampling of data

## FOR DOCUMENTATION:
#   - Clearly describe how chanset & refset works, ie. chanset = per chan, refset = ALL chans

class pipeline:
        
    """Contains specific information and allows the application of methods of 
    analysis, associated with a dataset. 

    Parameters
    ----------
    indir : str 
        name of the root level directory containing the BIDS organised data
        
    outfile : bool / str
        whether to save log of dataset audit to file. If False (default) - does
        not save any log file. If True - saves a log under the filepath 
        /derivatives/seapipe/audit/audit.csv. Else if a string to a filepath, 
        it will save the log under the filepath indicated in the string.
        
    filetype : str
        extension of file to search for. Default is '.edf' - but according to 
        BIDS convention other filetypes can be '.vhdr', '.vmrk', '.eeg' or '.set'

    Attributes
    ----------
    rootpath : str
        name of the root level directory
    datapath : str
        name of the directory containing the raw data (recordings and annotations)
    outpath : str
        name of the directory containing the output (analysed) data

    """
        
    def __init__(self, indir, tracking = False, outfile = False, 
                       filetype = '.edf'):
        
        self.rootpath = indir
        self.datapath = indir + '/DATA'
        self.outpath = indir + '/derivatives'
        if not path.exists(self.outpath):
            mkdir(self.outpath)
        self.outfile = outfile
        if not path.exists(f'{self.outpath}/audit'):
            mkdir(f'{self.outpath}/audit')
        self.tracking = {}
        self.audit_init = check_dataset(self.rootpath, self.outfile, filetype,
                                        tracking)
        self.track(subs = 'all', ses = 'all', 
                   step = ['staging','spindle','slowwave','pac','sync','psa'],
                   show = False, log = False)
        
    
    #--------------------------------------------------------------------------
    '''
    MISCELLANEOUS FUNCTIONS
    
    audit -> Audits dataset structure for compatibility with seapipe analysis.
    
    list_dataset ->  Intended to walk from root directory through participant 
                        folders and list all participants and their files.
    
    track -> Tracks what seapipe processing or functions have already been applied
                to a dataset, with information on which channels and parameters 
                have been used.
                
    make_bids (beta) -> Transforms data from (some) data structures into the 
                            correct BIDS format compatible with use for seapipe.
                            
    extract_channels -> Extracts and lists which channels exist in the dataset. 

    load_stages -> Extracts stages from the BIDS formatted dataset, in which
                        staging has been listed in a file *acq-PSGScoring_events.tsv
    
    '''    
        
        
    def audit(self, outfile = False, tracking = False, filetype = '.edf'):
        
        ''' Audits the dataset for BIDS compatibility.
            Includes option to save the audit to an output file.
        '''
        
        # Create audit directory
        out_dir = f'{self.outpath}/audit'
        if not path.exists(out_dir):
            mkdir(out_dir)
            
        if not outfile and not self.outfile:
            logger = create_logger("Audit")
            logger.propagate = False
            self.audit_update = check_dataset(self.rootpath, False, filetype, 
                                                             tracking, logger)
        else:
            if not outfile:
                outfile = self.outfile
            out = f'{out_dir}/{outfile}'
            if path.exists(out):
                remove(out)
            logger = create_logger_outfile(outfile, name = 'Audit')
            logger.propagate = False
            self.audit_update = check_dataset(self.rootpath, out, filetype, 
                                                             tracking, logger)
            
        logger.info('')
        logger.info(self.audit_update)
        
        
    def list_dataset(self, outfile=False): 
        
        """Prints out all the files inside the directory <in_dir> along with the
        directories 1 and 2 levels above containing the files. You can specify 
        an optional output filename that will contain the printout.
        """

        if not outfile and not self.outfile:
            logger = create_logger('Audit')  
        else:
            if not outfile:
                outfile = self.outfile
            out_dir = f'{self.outpath}/audit'
            if not path.exists(out_dir):
                mkdir(out_dir)
            out = f'{out_dir}/{outfile}'
            if path.exists(out):
                remove(out)
            logger = create_logger_outfile(out, name='Audit')

        logger.propagate = False
        
        logger.info("")
        logger.info("")
        for dirPath, dirNames, fileNames in walk(self.datapath):
            try:
                fileNames.remove('.DS_Store')
            except(ValueError):
                pass
            
            if fileNames or dirPath.split('/')[-1]=='eeg':
                dir1 = dirPath.split('/')[-3]
                dir2 = dirPath.split('/')[-2]
                dir3 = dirPath.split('/')[-1]
                logger.info(f"Directory: {dir1}/{dir2}/{dir3}")
                logger.info(f"Files; {fileNames}")
                logger.info('-' * 10)

    
    def track(self, subs = 'all', ses = 'all', step = None, chan = None, 
                    stage = None, outfile = False, show = True, log = True):
        
        ## Set up logging
        logger = create_logger('Tracking')
        logger.info('')
        
        ## Set tracking variable
        if self.tracking:
            tracking = self.tracking
        else:
            tracking = {}
        
        ## Track sessions  
        if not isinstance(subs, list) and subs == 'all':
            subs = [x for x in listdir(self.datapath) if '.' not in x]
        elif not isinstance(subs, list):
            
            subs = read_tracking_sheet(self.rootpath, logger)
            subs = subs['sub'].drop_duplicates().tolist()
        subs.sort()
        
        # Tracking
        tracking['ses'] = {}
        for sub in subs:
            try:
                tracking['ses'][sub] = [x for x in listdir(f'{self.datapath}/{sub}') 
                                    if '.' not in x]
            except:
                logger.warning(f'No sessions found for {sub}')
                tracking['ses'][sub] = ['-']
            
        # Dataframe
        df = DataFrame(data=None, dtype=object)
        df.index = subs
        df['ses'] = '-'
        for x in df.index:
            df.loc[x,'ses'] = tracking['ses'][x]
        
        # Loop through other steps
        if step: 
            df, tracking = track_processing(self, step, subs, tracking, df, chan, 
                                                  stage, show, log)

        # Update tracking
        try:
            self.tracking = self.tracking | tracking
        except:
            self.tracking = {**self.tracking, **tracking}
        
        if show:
            logger.info('')
            logger.info(df)
        if outfile:
            df.to_csv(f'{self.outpath}/audit/{outfile}')

        return   

    def make_bids(self, subs = 'all', origin = 'SCN'):
        make_bids(self.datapath, subs, origin)
        
    def extract_channels(self, exclude = None):
        extract_channels(self.datapath, exclude)
    
    def load_stages(self, xml_dir = None, subs = 'all', sessions = 'all', 
                          filetype = '.edf', stage_key = None, 
                          outfile = True):
        '''
            Extracts stages from the BIDS formatted dataset, in which
            staging has been listed in a file *acq-PSGScoring_events.tsv, and
            saves the information in an annotations file
        '''
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/load_sleep_stages_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Load sleep stages')
            logger.info('')
            logger.info(f"-------------- New call of 'Load sleep stages' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Load sleep stages')
        else:
            logger = create_logger('Load sleep stages')
        logger.info('')
        
        # Set xml_dir
        if not xml_dir:
            xml_dir = self.outpath + '/staging_manual'
        
        # Load stages
        flag = load_stages(self.datapath, xml_dir, subs, sessions, filetype, 
                           stage_key)
        
        # Log finish 
        if flag > 0:
            logger.warning(f"'load_stages' finished with {flag} WARNINGS. See log for detail.")
        else:
            logger.debug(f"'load_stages' finished without error.")
    #--------------------------------------------------------------------------
    '''
    ANALYSIS FUNCTIONS
    
    power_spectrum -> performs power spectral analysis.
                    
    
    '''    
    
    
    def power_spectrum(self, xml_dir = None, out_dir = None, 
                             subs = 'all', sessions = 'all', chan = None, 
                             ref_chan = None, grp_name = 'eeg', rater = None, 
                             stage = ['NREM1','NREM2','NREM3', 'REM'], 
                             cycle_idx = None, concat_cycle = True, 
                             concat_stage = False, general_opts = None, 
                             frequency_opts = None, filter_opts = None, 
                             epoch_opts = None, event_opts = None, 
                             norm = None, norm_opts = None, 
                             filetype = '.edf', outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_power_spectrum_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum')
            logger.info('')
            logger.info(f"-------------- New call of 'Power spectrum' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Power spectrum')
        else:
            logger = create_logger('Power spectrum')
        logger.info('')

        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'   
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/powerspectrum' 
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {xml_dir}')
        
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        
        # Set default parameters
        if not general_opts:
            general_opts = default_general_opts()
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        if not norm_opts:
            norm_opts = default_norm_opts()
        if not filter_opts:
            filter_opts = default_filter_opts()    
        frequency_opts['frequency'] = (filter_opts['highpass'], filter_opts['lowpass'])
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),
               int(epoch_opts['concat_signal']),
               int(event_opts['concat_events']),
               )
        
        # Set suffix for output filename
        if not general_opts['suffix']:
            general_opts['suffix'] = f"{frequency_opts['frequency'][0]}-{frequency_opts['frequency'][1]}Hz"
        
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
        else:   
            spectrum = Spectrum(in_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                                grp_name, stage, cat, rater, cycle_idx, subs, 
                                sessions, self.tracking)
                    
            spectrum.powerspec_it(general_opts, frequency_opts, filter_opts, 
                                  epoch_opts, event_opts, norm, norm_opts, 
                                  filetype, logger) 
            
            try:
                self.tracking = self.tracking | spectrum.tracking
            except:
                self.tracking = {**self.tracking, **spectrum.tracking}
            
        return 
        

    #--------------------------------------------------------------------------
    '''
    SLEEP EVENTS DETECTIONS
    
    sleep_staging
    detect_artefacts,
    detect_spectral_peaks,
    detect_slow_oscillations,
    detect_spindles,
    
    
    '''
    def detect_sleep_stages(self, xml_dir = None, out_dir = None, 
                                  subs = 'all', sessions = 'all', filetype = '.edf', 
                                  method = 'Vallat2021', qual_thresh = 0.5,
                                  eeg_chan = None, ref_chan = None, 
                                  eog_chan = None, emg_chan = None, 
                                  rater = None, invert = False, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_sleep_stages_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect sleep stages')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect sleep stages' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect sleep stages')
        else:
            logger = create_logger('Detect sleep stages')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'  
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/staging'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
        
        # Set channels
        eeg_chan, ref_chan = check_chans(self.rootpath, eeg_chan, ref_chan, logger)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('')
            logger.info('Check documentation for how to set up staging data: https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Sleep stage detection finished with ERRORS. See log for details.')
            return
    
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('')
            logger.info('Check documentation for how to set up staging data: https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Sleep stage detection finished with ERRORS. See log for details.')
        else:   
           stages = seabass(in_dir, xml_dir, out_dir, log_dir, eeg_chan, 
                            ref_chan, eog_chan, emg_chan, rater, subs, sessions, 
                            self.tracking) 
           stages.detect_stages(method, qual_thresh, invert, filetype, 
                                outfile, logger)
           try:
               self.tracking = self.tracking | stages.tracking
           except:
               self.tracking = {**self.tracking, **stages.tracking}
        return
    
    
    def detect_artefacts(self, xml_dir = None, out_dir = None, 
                               subs = 'all', sessions = 'all', filetype = '.edf', 
                               method = 'yasa_std', win_size = 5,
                               eeg_chans = None, ref_chan = None, 
                               eog_chan = None, emg_chan = None, 
                               rater = None, invert = False, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_artefacts_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect artefacts')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect artefacts' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect artefacts')
        else:
            logger = create_logger('Detect artefacts')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'   
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/staging'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
        
        # Set channels
        eeg_chans, ref_chan = check_chans(self.rootpath, eeg_chans, ref_chan, logger)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('')
            logger.info('Check documentation for how to set up staging data: https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Sleep stage detection finished with ERRORS. See log for details.')
            return
    
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('')
            logger.info('Check documentation for how to set up staging data: https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Sleep stage detection finished with ERRORS. See log for details.')
        else:   
           artefacts = SAND(in_dir, xml_dir, out_dir, log_dir, eeg_chans, 
                            ref_chan, eog_chan, emg_chan, rater, subs, sessions, 
                            self.tracking) 
           artefacts.detect_artefacts(method, invert, filetype, win_size, logger)
       
           try:
               self.tracking = self.tracking | artefacts.tracking
           except:
               self.tracking = {**self.tracking, **artefacts.tracking}
        return
        
        
    
    def detect_spectral_peaks(self, xml_dir = None, out_dir = None, 
                                    subs = 'all', sessions = 'all', chan = None, 
                                    ref_chan = None, grp_name = 'eeg', 
                                    rater = None, frequency = (9,16), 
                                    stage = ['NREM2','NREM3'], cycle_idx = None,
                                    concat_cycle = True, concat_stage = False,
                                    general_opts = None, frequency_opts = None,
                                    filter_opts = None, epoch_opts = None, 
                                    event_opts = None, fooof_opts = None, 
                                    filetype = '.edf', suffix = None, 
                                    outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_specparams_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect spectral peaks')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spectral peaks' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect spectral peaks')
        else:
            logger = create_logger('Detect spectral peaks')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'  
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/fooof' 
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {xml_dir}')
            
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()    
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)
        
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
        else:   
            spectrum = Spectrum(in_dir, xml_dir, out_dir, log_dir, chan, 
                                ref_chan, grp_name, stage, cat, rater, 
                                cycle_idx, subs, sessions, self.tracking)
            
            if not general_opts:
                general_opts = default_general_opts()
            if not frequency_opts:
                frequency_opts = default_frequency_opts()
            if not filter_opts:
                filter_opts = default_filter_opts()
            if not epoch_opts:
                epoch_opts = default_epoch_opts()  
            if not event_opts:
                event_opts = default_event_opts()
            if not fooof_opts:
                fooof_opts = default_fooof_opts() 
                
            fooof_opts['bands_fooof'] = [frequency]
            
            # Set suffix for output filename
            if not suffix:
                general_opts['suffix'] = f'{frequency[0]}-{frequency[1]}Hz'
            
            spectrum.fooof_it(general_opts, frequency_opts, filter_opts, 
                              epoch_opts, event_opts, fooof_opts, 
                              filetype, logger)  
                
        return 
    
    
    def detect_slow_oscillations(self, xml_dir=None, out_dir=None, subs='all', 
                                       sessions='all', filetype='.edf', 
                                       method = ['Staresina2015'], chan=None,
                                       ref_chan=None, rater=None, grp_name='eeg', 
                                       stage = ['NREM2','NREM3'], cycle_idx=None, 
                                       duration=(0.2, 2), invert = None,
                                       reject_artf = ['Artefact', 'Arou', 'Arousal'],
                                       average_channels = False, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_slowosc_{evt_out}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect slow oscillations')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect slow oscillations' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect slow oscillations')
        else:
            logger = create_logger('Detect slow oscillations')
        
        logger.info('')
        logger.debug("Commencing SO detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'   
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/slowwave'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('SO detection finished with ERRORS. See log for details.')
            return
            
        # Format concatenation
        cat = (1,1,1,1)
    
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('SO detection finished with ERRORS. See log for details.')
        else:   
           SO = seasnakes(in_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                            grp_name, stage, rater, subs, sessions, 
                            self.tracking, reject_artf) 
           SO.detect_slowosc(method, cat, cycle_idx, duration, 
                                  average_channels, invert, filetype, logger)
           try:
               self.tracking = self.tracking | SO.tracking
           except:
               self.tracking = {**self.tracking, **SO.tracking}
        return
    
    
    def detect_spindles(self, xml_dir = None, out_dir = None, subs = 'all', 
                              sessions = 'all', filetype = '.edf', 
                              method = ['Moelle2011'], chan = None, 
                              ref_chan = None, rater = None, 
                              stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                              cycle_idx = None, concat_cycle = True, 
                              frequency = None, adap_bands = 'Fixed', 
                              adap_bw = 4, duration =( 0.5, 3),
                              reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                              outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_spindles_{evt_out}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spindles' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles')
        else:
            logger = create_logger('Detect spindles')
            

        logger.info('')
        logger.debug("Commencing spindle detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging' 
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            for met in method:
                out_dir = select_output_dirs(self.outpath, out_dir, met)  
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            return
        elif isinstance(ref_chan, str):
            return
        
        # Format concatenation
        if concat_cycle == True:
            cat = (1,0,1,1)
        else:
            cat = (0,0,1,1)
            if not cycle_idx:
                logger.error("'concat_cycle' is set to false, but 'cycle_idx' = None. Set cycle_idx to a list of integers to use cycles properly.")
                logger.info("Check documentation for how to mark and use sleep cycles:")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                return
                
        # Check for adapted bands
        if adap_bands == 'Fixed':
            logger.debug('Detection using FIXED frequency bands has been selected (adap_bands = Fixed)')
            if not frequency:
                frequency = (11,16)
        elif adap_bands == 'Manual':
            logger.debug('Detection using ADAPTED (user-provided) frequency bands has been selected (adap_bands = Manual)')
            logger.debug(f"Checking for spectral peaks in {self.rootpath}/'tracking.tsv' ")
            flag = check_adap_bands(self.rootpath, subs, sessions, chan, logger)
            if flag == 'error':
                logger.critical('Spindle detection finished with ERRORS. See log for details.')
                return
            elif flag == 'review':
                logger.info('')
                logger.warning(f"Some spectral peak entries in 'tracking.tsv' are inconsistent or missing. In these cases, detection will revert to fixed bands: {frequency[0]}-{frequency[1]}Hz")
                logger.info('')
        elif adap_bands == 'Auto': 
            if not frequency:
                frequency = (9,16)           
            logger.debug('Detection using ADAPTED (automatic) frequency bands has been selected (adap_bands = Auto)')
            self.track(subs, sessions, step = 'fooof', show = False, log = False)
            if not type(chan) == type(DataFrame()):
                logger.critical("For adap_bands = Auto, the argument 'chan' must be 'None' and specfied in 'tracking.csv'")
                logger.critical('Spindle detection finished with ERRORS. See log for details.')
                return
            else:
                flag, pk_chan, pk_sub, pk_ses = check_fooof(self, frequency, 
                                                                  chan, ref_chan, 
                                                                  stage, 
                                                                  cat,
                                                                  cycle_idx, 
                                                                  logger)
            if flag == 'error':
                logger.critical('Error in reading channel names, check tracking sheet.')
                logger.info("Check documentation for how to set up channel names in tracking.tsv':")
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                logger.critical('Spindle detection finished with ERRORS. See log for details.')
                return
            elif flag == 'review':
                logger.debug('Spectral peaks have not been found for all subs, analysing the spectral parameters prior to spindle detection..')
                for (sub,ses) in zip(pk_sub,pk_ses):
                    self.detect_spectral_peaks(subs = [sub], 
                                           sessions = [ses], 
                                           chan = pk_chan, 
                                           frequency = frequency,
                                           stage = stage, cycle_idx = cycle_idx,
                                           concat_cycle=concat_cycle, 
                                           concat_stage=True)
        # Check annotations directory exists, run detection
        self.track(step='fooof', show = False, log = False)
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Spindle detection finished with ERRORS. See log for details.')
        else:   
           spindle = whales(self.rootpath, in_dir, xml_dir, out_dir, log_dir, 
                            chan, ref_chan, grp_name, stage, frequency, rater, 
                            subs, sessions, reject_artf, self.tracking) 
           spindle.whale_it(method, cat, cycle_idx, adap_bands, adap_bw, 
                            duration, filetype, logger)
           try:
               self.tracking = self.tracking | spindle.tracking
           except:
               self.tracking = {**self.tracking, **spindle.tracking}
        return
    
    
    def whales(self, xml_dir = None, out_dir = None, subs = 'all', 
                     sessions = 'all', filetype = '.edf', 
                     method = ['Moelle2011', 'Ray2015'], evt_out = 'spindle',
                     merge_type = 'consensus', weights = None,
                     chan = None, ref_chan = None, rater = None, 
                     stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                     cycle_idx = None, s_freq = None, keyword = None, 
                     duration =(0.5, 3),
                     reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                     outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_spindles_WHALES_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles (WHALES)')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect spindles (WHALES)' evoked at {now} --------------")
       
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect spindles (WHALES)')
        else:
            logger = create_logger('Detect spindles (WHALES)')
            
        logger.info('')
        logger.debug("Commencing spindle optimisation pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        xml_dir = select_input_dirs(self.outpath, xml_dir, 'spindle') 
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = xml_dir  
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        if merge_type == 'consensus':
            cs_thresh = 0.5
        elif merge_type == 'addition':
            cs_thresh = 0.01
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
        
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            return
        elif isinstance(ref_chan, str):
            return
        
        spindle = whales(self.rootpath, in_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                         grp_name, stage, frequency = None, rater = rater, 
                         subs = subs, sessions = sessions, 
                         reject_artf = reject_artf, tracking = self.tracking) 
        spindle.whales(method, merge_type, chan, rater, stage, ref_chan, grp_name, 
                       keyword, cs_thresh, s_freq, duration, evt_out, weights,
                       filetype, logger)
    
    
    def detect_rems(self, xml_dir = None, out_dir = None, subs = 'all', 
                          sessions = 'all', filetype = '.edf', 
                          method = ['YASA'], chan = None,
                          ref_chan = None, rater = None, grp_name = 'eeg', 
                          stage = ['REM'], cycle_idx = None, 
                          amplitude = (50, 325), duration = (0.3, 1.5),
                          reject_artf = ['Artefact', 'Arou', 'Arousal'],
                          average_channels = False, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs'
        if outfile == True:
            evt_out = '_'.join(method)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            logfile = f'{log_dir}/detect_rems_{evt_out}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Detect eye movements (REMS)')
            logger.info('')
            logger.info(f"-------------- New call of 'Detect rapid eye movements' evoked at {now} --------------")
            
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Detect Detect REMS')
        else:
            logger = create_logger('Detect Detect REMS')
        
        logger.info('')
        logger.debug("Commencing REMS detection pipeline.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'   
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/rems'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
    
        # Check annotations directory exists, run detection
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Sleep staging has not been run or hasn't been converted correctly.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('REMS detection finished with ERRORS. See log for details.')
        else:   
           REMS = remora(in_dir, xml_dir, out_dir, log_dir, chan, ref_chan, 
                         rater, grp_name, reject_artf, subs, sessions)
           
           REMS.detect_rems(method, amplitude, duration, stage, cycle_idx, 
                            filetype, logger)
           try:
               self.tracking = self.tracking | REMS.tracking
           except:
               self.tracking = {**self.tracking, **REMS.tracking}
        return
    
    #--------------------------------------------------------------------------
    '''
    PLOTTING.
    
    event_spectrogram ->
    
    
    '''
    
    
    def spectrogram(self, xml_dir = None, out_dir = None, subs = 'all', 
                          sessions = 'all', filetype = '.edf', chan = None, 
                          ref_chan = None, rater = None, stage = None, 
                          grp_name = 'eeg', cycle_idx = None, 
                          concat_stage = False, concat_cycle = True, 
                          evt_type = None, buffer = 0, invert = None, 
                          filter_opts = None, progress=True, outfile=False):
        
        # Set up logging
        logger = create_logger('Event spectrogram')
        logger.info('')
        logger.debug("Creating spectrogram of events.")
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        log_dir = self.outpath + '/audit/logs/'
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not xml_dir:
            xml_dir = f'{self.outpath}/staging'  
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not out_dir:
            out_dir = f'{self.outpath}/spindle'    
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output annotations being saved to: {out_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, chan, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        if not filter_opts:
            filter_opts = default_filter_opts()
        
        
        if not evt_type:
            logger.warning('No event type (evt_type) has been specified. Spectrogram will be run on epochs instead. This may take some time...')
        
        event_spectrogram(self, in_dir, xml_dir, out_dir, subs, sessions, stage, 
                                cycle_idx, chan, ref_chan, rater, grp_name, 
                                evt_type, buffer, invert, cat, filter_opts,  
                                outfile, filetype, progress, self.tracking)
        
        return
    
    
    #--------------------------------------------------------------------------
    '''
    PHASE AMPLITUDE COUPLING.
    
    mean_amps -> runs Phase Amplitude Coupling analyses on sleep EEG data. 
    
    
    '''    
    def pac(self, xml_dir = None, out_dir = None, subs = 'all', sessions = 'all', 
                  filetype = '.edf', chan = None, ref_chan = None, rater = None, 
                  grp_name = 'eeg', stage = ['NREM2','NREM3'], concat_stage = True, 
                  cycle_idx = None, concat_cycle = True,  
                  method = 'MI', surrogate = 'Time lag', correction = 'Z-score',
                  adap_bands_phase = 'Fixed', frequency_phase = (0.5, 1.25), 
                  adap_bands_amplitude = 'Fixed', frequency_amplitude = (11, 16),
                  adap_bw = 4, min_dur = 1, nbins = 18, invert = None,
                  frequency_opts = None, filter_opts = None, epoch_opts = None, 
                  evt_name = None, event_opts = None, 
                  reject_artf = ['Artefact', 'Arou', 'Arousal'], 
                  progress = True, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H%M%S")
            pha = f'{frequency_phase[0]}-{frequency_phase[1]}'
            amp = f'{frequency_amplitude[0]}-{frequency_amplitude[1]}'
            logfile = f'{log_dir}/pac_{pha}_{amp}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Phase-amplitude coupling')
            logger.info('')
            logger.info(f"-------------- New call of 'Phase-amplitude coupling' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Phase-amplitude coupling')
        else:
            logger = create_logger('Phase-amplitude coupling')
        logger = create_logger('Phase-amplitude coupling')
        logger.info('')
        
        # Set input/output directories
        in_dir = self.datapath
        log_dir = self.outpath + '/audit/logs/'
        if not path.exists(log_dir):
            mkdir(log_dir) 
        if not xml_dir:
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name) 
        logger.debug(f'Input annotations being read from: {xml_dir}')
        
        # Check subs
        if not subs:
            tracking = read_tracking_sheet(self.rootpath, logger)
            subs = [x for x in list(set(tracking['sub']))]
            subs.sort()
        if not sessions:
            sessions = read_tracking_sheet(self.rootpath, logger)
            
        # Set channels
        chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
        if not isinstance(chan, DataFrame) and not isinstance(chan, list):
            return
        elif isinstance(ref_chan, str):
            return
        
        # Set PAC methods
        idpac = pac_method(method, surrogate, correction)
        
        # Set default parameters
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        if not filter_opts:
            filter_opts = default_filter_opts()   
        filter_opts['bandpass'] = False
        
        # Check inversion
        if invert == None:
            invert = check_chans(self.rootpath, None, False, logger)
        elif type(invert) != bool:
            logger.critical(f"The argument 'invert' must be set to either: 'True', 'False' or 'None'; but it was set as {invert}.")
            logger.info('Check documentation for how to set up staging data:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            logger.critical('Phase amplitude coupling finished with ERRORS. See log for details.')
            return
        
        # Check whether event based or continuous
        if evt_name: #OCTOPUS
            if not out_dir:
                out_dir = f'{self.outpath}/event_pac'  
            if not path.exists(out_dir):
                mkdir(out_dir)
            logger.debug(f'Output being saved to: {xml_dir}')
                
            cat = (int(concat_cycle),int(concat_stage),0,0)
            Octopus = octopus(self.rootpath, in_dir, xml_dir, out_dir, log_dir, 
                              chan, ref_chan, grp_name, stage, rater, 
                              subs, sessions, reject_artf,
                              self.tracking)
            
            Octopus.pac_it(cycle_idx, cat, nbins, filter_opts, epoch_opts, 
                           frequency_opts, event_opts, filetype, idpac, evt_name, 
                           min_dur, adap_bands_phase, frequency_phase, 
                           adap_bands_amplitude, frequency_amplitude, 
                           adap_bw, invert, progress, outfile)
        else: #PACATS
            if not out_dir:
                out_dir = f'{self.outpath}/pac'  
            if not path.exists(out_dir):
                mkdir(out_dir)
            logger.debug(f'Output being saved to: {xml_dir}')
            cat = (int(concat_cycle),int(concat_stage),1,1)
            Pacats = pacats(self.rootpath, in_dir, xml_dir, out_dir, log_dir, 
                            chan, ref_chan, grp_name, stage, rater, subs, sessions, 
                            reject_artf, self.tracking)
            Pacats.pac_it(cycle_idx, cat, nbins, filter_opts, epoch_opts, 
                           frequency_opts, filetype, idpac, 
                           min_dur, adap_bands_phase, frequency_phase, 
                           adap_bands_amplitude, frequency_amplitude,
                           adap_bw, invert, progress, logger)
            
                          
        return
    
    #--------------------------------------------------------------------------
    '''
    DATASET CREATION.
    
    export_macro_stats -> Exports sleep macroarchitecture per participant into 
                            the corresponding folder in output directory 'staging' 
    
    macro_dataset -> Creates a cohort dataset of sleep macroarchitecture and saves
                        it to a single .csv file in output directory 'dataset'
    
    export_eventparams -> Exports descriptives for sleep events per participant into 
                            the corresponding folder in output directory 'staging'
    
    event_dataset -> Creates a cohort dataset of sleep events descriptives and saves
                        it to a single .csv file in output directory 'dataset'
    
    '''    
    
    def export_macro_stats(self, xml_dir = None, out_dir = None, 
                                 subs = 'all', sessions = 'all', 
                                 times = None, rater = None, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{log_dir}/export_sleep_macro_stats_{today}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
            logger.info('')
            logger.info(f"-------------- New call of 'Export macro stats' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
        else:
            logger = create_logger('Export macro stats')

        
        # Set input/output directories
        if not path.exists(log_dir):
            mkdir(log_dir)
        xml_dir = select_input_dirs(self.outpath, xml_dir, 'macro')
        logger.debug(f'Input annotations being read from: {xml_dir}')
        out_dir = select_output_dirs(self.outpath, out_dir, 'macro')
        logger.debug(f'Output being saved to: {out_dir}')
        
        # Set channels
        times, ref_chan = check_chans(self.rootpath, None, True, logger)
        
        self.track(subs=subs, ses=sessions, 
                   step=['staging'],
                   show=False, log=True)
        
        sleepstats.export_sleepstats(xml_dir, out_dir, subs, sessions, 
                                     rater, times, log_dir, logger)
        return
    
    def macro_dataset(self, xml_dir = None, out_dir = None, 
                      subs = 'all', sessions = 'all', outfile = True):
         
         # Set input/output directories
         log_dir = self.outpath + '/audit/logs/'
         if outfile == True:
             today = date.today().strftime("%Y%m%d")
             now = datetime.now().strftime("%H:%M:%S")
             logfile = f'{log_dir}/export_sleep_macro_stats_{today}_log.txt'
             logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
             logger.info('')
             logger.info(f"-------------- New call of 'Macro dataset' evoked at {now} --------------")
         elif outfile:
             logfile = f'{log_dir}/{outfile}'
             logger = create_logger_outfile(logfile=logfile, name='Export macro stats')
         else:
             logger = create_logger('Export macro stats')
         if not path.exists(log_dir):
             mkdir(log_dir)
         if not path.exists(self.outpath + '/datasets/'):
             mkdir(self.outpath + '/datasets/')
         out_dir = self.outpath + '/datasets/macro/'
         
         xml_dir = select_input_dirs(self.outpath, xml_dir, 'macro')
         logger.debug(f'Input annotations being read from: {xml_dir}')
         out_dir = select_output_dirs(self.outpath, out_dir, 'macro')
         logger.debug(f'Output being save to: {out_dir}')
         
         sleepstats.sleepstats_from_csvs(xml_dir, out_dir,   
                                 subs, sessions, log_dir, logger)
         return
    
    def export_eventparams(self, evt_name, frequency = None,
                                 xml_dir = None, out_dir = None, subs = 'all', 
                                 sessions = 'all', chan = None, ref_chan = None, 
                                 stage = ['NREM2','NREM3'], grp_name = 'eeg', 
                                 rater = None, cycle_idx = None, 
                                 concat_cycle = True, concat_stage = False, 
                                 keyword = None, segs = None,  
                                 adap_bands = 'Fixed',  
                                 adap_bw = 4, params = 'all', epoch_dur = 30, 
                                 average_channels = False, outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            evt_out = '_'.join(evt_name)
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{log_dir}/export_params_{evt_out}_subs-{subs}_ses-{sessions}_{today}_{now}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Export params')
            logger.info('')
            logger.debug(f"-------------- New call of 'Export params' evoked at {now} --------------")
        elif outfile:
            logfile = f'{log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Export params')
        else:
            logger = create_logger('Export params')
        
        # Set input/output directories
        in_dir = self.datapath
        if not path.exists(log_dir):
            mkdir(log_dir)
        
        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            logger.error(TypeError(f"'evt_name' can only be a str or a list, but {type(evt_name)} was passed."))
            return
        for evt_name in evts:
            out_dir = select_output_dirs(self.outpath, out_dir, evt_name)
            logger.debug(f'Output being save to: {out_dir}')
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name)
            logger.debug(f'Input annotations being read from: {xml_dir}')
            
            
            # Check annotations directory exists
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Event detection has not been run or an incorrect event type has been selected.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
    
            if adap_bands in ['Auto','Manual']:
                evt_name = f'{evt_name}_adap'
                self.track(step='fooof', show = False, log = False)
                peaks = check_chans(self.rootpath, None, False, logger)
            else:
                peaks = None
            
            # Set channels
            chan, ref_chan = check_chans(self.rootpath, chan, ref_chan, logger)
            if average_channels:
                Ngo = {'run':True}
            else:
                Ngo = {'run':False}
            
            # Format concatenation
            cat = (int(concat_cycle),int(concat_stage),1,1)
            
            # Run line
            fish = FISH(self.rootpath, in_dir, xml_dir, out_dir, log_dir, chan, ref_chan, grp_name, 
                              stage, rater, subs, sessions, self.tracking) 
            fish.line(keyword, evt_name, cat, segs, cycle_idx, frequency, adap_bands, 
                      peaks, adap_bw, params, epoch_dur, Ngo, logger)
        return
    
    
    def event_dataset(self, chan, evt_name, xml_dir = None, out_dir = None, 
                            subs = 'all', sessions = 'all', 
                            stage = ['NREM2','NREM3'], concat_stage = False, 
                            concat_cycle = True, cycle_idx = None, 
                            grp_name = 'eeg', adap_bands = 'Fixed',  
                            params = 'all', outfile = True):
        
        # Set up logging
        log_dir = self.outpath + '/audit/logs/'
        if outfile == True:
            today = date.today().strftime("%Y%m%d")
            now = datetime.now().strftime("%H:%M:%S")
            logfile = f'{self.log_dir}/event_dataset_{evt_name}_{today}_log.txt'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
            logger.info('')
            logger.info(f'-------------- New call evoked at {now} --------------')
        elif outfile:
            logfile = f'{self.log_dir}/{outfile}'
            logger = create_logger_outfile(logfile=logfile, name='Event dataset')
        else:
            logger = create_logger('Event dataset')
        logger = create_logger('Event dataset')
        
        # Force evt_name into list, and loop through events    
        if isinstance(evt_name, str):
            evts = [evt_name]
        elif isinstance(evt_name, list):
            evts = evt_name
        else:
            logger.error(TypeError(f"'evt_name' can only be a str or a list of str, but {type(evt_name)} was passed."))
            logger.info('Check documentation for how to create an event_dataset:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        for evt_name in evts:
            # Append 'adap' after event name if adapted bands were used
            if adap_bands in ['Auto', 'Manual']:
                evt_name = f'{evt_name}_adap'
                self.track(step='fooof', show = False, log = False)
            
            # Set input/output directories
            in_dir = self.datapath
            if not path.exists(log_dir):
                mkdir(log_dir)
            if not out_dir:    
                if not path.exists(self.outpath + '/datasets/'):
                    mkdir(self.outpath + '/datasets/')
                out_dir = self.outpath + f'/datasets/{evt_name}'
            if not path.exists(out_dir):
                mkdir(out_dir)
            logger.debug(f'Output being saved to: {xml_dir}')
            
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name)
            logger.debug(f'Input annotations being read from: {xml_dir}')
            if not path.exists(xml_dir):
                logger.info('')
                logger.critical(f"{xml_dir} doesn't exist. Event detection has not been run or an incorrect event type has been selected.")
                logger.info('Check documentation for how to run a pipeline:')
                logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
                logger.info('-' * 10)
                return
            
            # Format concatenation
            cat = (int(concat_cycle),int(concat_stage),1,1)
            
            # Format chan
            if isinstance(chan, str):
                chan = [chan]
            
        
            fish = FISH(self.rootpath, in_dir, xml_dir, out_dir, log_dir, chan, None, grp_name, 
                              stage, subs = subs, sessions = sessions) 
            fish.net(chan, evt_name, adap_bands, params,  cat, cycle_idx, logger)
        
        return
    
    def pac_dataset(self, chan, evt_name = None, subs = 'all', sessions = 'all',
                          xml_dir = None, out_dir = None,  stage = None, 
                          concat_stage = False, concat_cycle = True, 
                          cycle_idx = None, grp_name = 'eeg', 
                          adap_bands_phase = 'Fixed', frequency_phase = (0.5, 1.25), 
                          adap_bands_amplitude = 'Fixed', frequency_amplitude = (11, 16),  
                          params = 'all', outfile=True):
        
        # Set up logging
        logger = create_logger('PAC dataset')
        
        # Set input/output directories
        in_dir = self.datapath
        log_dir = self.outpath + '/audit/logs/'
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not out_dir:
            if not path.exists(self.outpath + '/datasets/'):
                mkdir(self.outpath + '/datasets/')
            out_dir = f'{self.outpath}/datasets/pac'
        if not path.exists(out_dir):
            mkdir(out_dir)
        logger.debug(f'Output being saved to: {xml_dir}')
        
        # Check if event-based or continuous PAC
        if isinstance(evt_name, str):
            evt = 'event_pac' 
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt)
            cat = (int(concat_cycle),int(concat_stage),1,1)
        elif evt_name == None:
            evt = 'pac' 
            xml_dir = select_input_dirs(self.outpath, xml_dir, evt)
            cat = (int(concat_cycle),int(concat_stage),0,0)
        else:
            logger.error(TypeError(f"'evt_name' can only be a str or NoneType, but {type(evt_name)} was passed."))
            logger.info('Check documentation for how to create a PAC summary dataset:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. PAC detection has not been run or an incorrect type has been selected.")
            logger.info('Check documentation for how to run a pipeline:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return

        

        # Format chan
        if isinstance(chan, str):
            chan = [chan]
        
        # Default stage
        if stage == None:
            stage = ['NREM2','NREM3']
            
        fish = FISH(self.rootpath, in_dir, xml_dir, out_dir, log_dir, chan, None, grp_name, 
                          stage, subs = subs, sessions = sessions) 
        fish.pac_summary(chan, evt_name, adap_bands_phase, frequency_phase, 
                              adap_bands_amplitude, frequency_amplitude,
                              params = 'all',
                              cat = cat, cycle_idx = None, outfile = True)
        
        return

    def powerspec_dataset(self, chan, xml_dir = None, out_dir = None, 
                                subs = 'all', sessions = 'all', 
                                stage = ['NREM1','NREM2','NREM3', 'REM'], 
                                concat_stage = False, concat_cycle = True, 
                                cycle_idx = None, grp_name = 'eeg', 
                                rater = None, params = 'all', 
                                general_opts = None, frequency_opts = None, 
                                filter_opts = None, epoch_opts = None, 
                                event_opts = None, outfile=True):
        
        # Set up logging
        logger = create_logger('Power spectrum dataset')
        
        # Set input/output directories
        in_dir = self.datapath
        log_dir = self.outpath + '/audit/logs/'
        if not path.exists(log_dir):
            mkdir(log_dir)
        if not out_dir:
            if not path.exists(self.outpath + '/datasets/'):
                mkdir(self.outpath + '/datasets/')
            out_dir = f'{self.outpath}/datasets/powerspectrum'
            if not path.exists(out_dir):
                mkdir(out_dir)
        logger.debug(f'Output being saved to: {xml_dir}')
        
        xml_dir = select_input_dirs(self.outpath, xml_dir, evt_name = 'powerspectrum')
        logger.debug(f'Input annotations being read from: {xml_dir}')
        if not path.exists(xml_dir):
            logger.info('')
            logger.critical(f"{xml_dir} doesn't exist. Event detection has not been run or an incorrect event type has been selected.")
            logger.info('Check documentation for how to run a pipeline:')
            logger.info('https://seapipe.readthedocs.io/en/latest/index.html')
            logger.info('-' * 10)
            return
        
        # Set default parameters
        if not general_opts:
            general_opts = default_general_opts()
        if not frequency_opts:
            frequency_opts = default_frequency_opts()
        if not epoch_opts:
            epoch_opts = default_epoch_opts()  
        if not event_opts:
            event_opts = default_event_opts()
        
        if not filter_opts:
            filter_opts = default_filter_opts()    
        frequency_opts['frequency'] = (filter_opts['highpass'], filter_opts['lowpass'])
        
        # Set suffix for output filename
        if not general_opts['suffix']:
            general_opts['suffix'] = f"{frequency_opts['frequency'][0]}-{frequency_opts['frequency'][1]}Hz"
        
        # Format chan
        if isinstance(chan,str):
            chan = [chan]
        
        # Format concatenation
        cat = (int(concat_cycle),int(concat_stage),1,1)

        spectrum = Spectrum(in_dir, xml_dir, out_dir, log_dir, chan, 
                            ref_chan = None, grp_name = grp_name, stage = stage, 
                            cat = cat, rater = rater, cycle_idx = cycle_idx, 
                            subs = subs, sessions = sessions) 
        spectrum.powerspec_summary(chan, general_opts, frequency_opts, filter_opts, 
                                   epoch_opts, event_opts, logger)
        
        return
    
        
        